import asyncio
import functools
import logging
import os
import platform
import sys
from typing import Optional

import click
import lockfile
import psutil
from click import INT, STRING
from lockfile.pidlockfile import PIDLockFile, read_pid_from_pidfile, remove_existing_pidfile
from pydantic import AnyHttpUrl

from giskard.cli_utils import common_options
from giskard.cli_utils import (
    create_pid_file_path,
    remove_stale_pid_file,
    run_daemon,
    get_log_path,
    tail,
    follow_file,
    validate_url,
)
from giskard.path_utils import run_dir
from giskard.settings import settings
from giskard.utils.analytics_collector import anonymize, analytics

logger = logging.getLogger(__name__)


@click.group("worker", help="ML Worker management", context_settings={"show_default": True})
def worker() -> None:
    """
    ML Worker management
    """


def start_stop_options(func):
    @click.option(
        "--url",
        "-u",
        type=STRING,
        default="http://localhost:19000",
        help="Remote Giskard server url",
        callback=validate_url,
    )
    @click.option(
        "--server",
        "-s",
        "is_server",
        is_flag=True,
        default=False,
        help="Server mode. Used by Giskard embedded ML Worker",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@worker.command("start")
@common_options
@start_stop_options
@click.option(
    "--key",
    "-k",
    "api_key",
    envvar="GSK_API_KEY",
    help="Giskard server API key",
)
@click.option(
    "--daemon",
    "-d",
    "is_daemon",
    is_flag=True,
    default=False,
    help="Should ML Worker be started as a Daemon in a background",
)
def start_command(url: AnyHttpUrl, is_server, api_key, is_daemon):
    """\b
    Start ML Worker.

    ML Worker can be started in 2 modes:

    - server: used by default by an ML Worker shipped by Giskard. ML Worker acts as a server that Giskard connects to.

    - client: ML Worker acts as a client and should connect to a running Giskard instance
        by specifying this instance's host and port.
    """
    analytics.track(
        "giskard-worker:start",
        {"is_server": is_server, "url": anonymize(url), "is_daemon": is_daemon},
    )
    api_key = initialize_api_key(api_key, is_server)
    _start_command(is_server, url, api_key, is_daemon)


def initialize_api_key(api_key, is_server):
    if is_server:
        return None
    if not api_key:
        api_key = click.prompt("Enter Giskard server API key", type=str)
    if "GSK_API_KEY" in os.environ:
        # delete API key environment variable so that it doesn't get leaked when the test code is executed
        del os.environ["GSK_API_KEY"]
    return api_key


def _start_command(is_server, url: AnyHttpUrl, api_key, is_daemon):
    from giskard.ml_worker.ml_worker import MLWorker

    start_msg = "Starting ML Worker"
    start_msg += " server" if is_server else " client"
    if is_daemon:
        start_msg += " daemon"
    logger.info(start_msg)
    logger.info(f"Python: {sys.executable} ({platform.python_version()})")
    logger.info(f"Giskard Home: {settings.home_dir}")
    pid_file_path = create_pid_file_path(is_server, url)
    pid_file = PIDLockFile(pid_file_path)
    remove_stale_pid_file(pid_file)

    ml_worker: Optional[MLWorker] = None
    try:
        pid_file.acquire()
        if is_daemon:
            # Releasing the lock because it will be re-acquired by a daemon process
            pid_file.release()
            # If on windows, throw error and exit
            if sys.platform == "win32":
                logger.error("Daemon mode is not supported on Windows.")
                return

            run_daemon(is_server, url, api_key)
        else:
            ml_worker = MLWorker(is_server, url, api_key)
            asyncio.get_event_loop().run_until_complete(ml_worker.start())
    except KeyboardInterrupt:
        logger.info("Exiting")
        if ml_worker:
            asyncio.get_event_loop().run_until_complete(ml_worker.stop())
    except lockfile.AlreadyLocked:
        existing_pid = read_pid_from_pidfile(pid_file_path)
        logger.warning(
            f"Another ML Worker {_ml_worker_description(is_server, url)} "
            f"is already running with PID: {existing_pid}. "
            "Not starting a new one. "
            'To stop a running worker for this instance execute: "giskard worker stop" or '
            '"giskard worker stop -a" to stop all running workers'
        )
    finally:
        if pid_file.i_am_locking():
            pid_file.release()


def _ml_worker_description(is_server, url):
    return "server" if is_server else f"client for {url}"


@worker.command("stop", help="Stop running ML Workers")
@common_options
@start_stop_options
@click.option("--all", "-a", "stop_all", is_flag=True, default=False, help="Stop all running ML Workers")
def stop_command(is_server, url, stop_all):
    import re

    analytics.track(
        "giskard-worker:stop",
        {"is_server": is_server, "url": anonymize(url), "stop_all": stop_all},
    )
    if stop_all:
        for pid_fname in os.listdir(run_dir):
            if not re.match(r"^ml-worker-.*\.pid$", pid_fname):
                continue
            _stop_pid_fname(pid_fname)
    else:
        _find_and_stop(is_server, url)


@worker.command("restart", help="Restart ML Worker")
@common_options
@start_stop_options
@click.option("--api-key", "-k", "api_key", help="Giskard server API key")
def restart_command(is_server, url, api_key):
    analytics.track(
        "giskard-worker:restart",
        {"is_server": is_server, "url": anonymize(url)},
    )
    api_key = initialize_api_key(api_key, is_server)

    _find_and_stop(is_server, url)
    _start_command(is_server, url, api_key, is_daemon=True)


def _stop_pid_fname(pid_fname):
    pid_file_path = str(run_dir / pid_fname)
    remove_stale_pid_file(PIDLockFile(pid_file_path))
    pid = read_pid_from_pidfile(pid_file_path)
    if pid:
        worker_process = psutil.Process(pid)
        worker_process.terminate()
        logger.info(f"Stopped ML Worker Daemon by PID: {pid}")
    remove_existing_pidfile(pid_file_path)


def _find_and_stop(is_server, url):
    pid_file_path = str(create_pid_file_path(is_server, url))
    remove_stale_pid_file(PIDLockFile(pid_file_path))
    pid = read_pid_from_pidfile(pid_file_path)
    logger.info("Stopping ML Worker Daemon")
    if pid:
        worker_process = psutil.Process(pid)
        worker_process.terminate()
        logger.info(f"Stopped ML Worker {_ml_worker_description(is_server, url)}")
    else:
        logger.info(f"ML Worker {_ml_worker_description(is_server, url)} is not running")
    remove_existing_pidfile(pid_file_path)


@worker.command("logs")
@common_options
@click.option(
    "--lines",
    "-n",
    type=INT,
    default=10,
    help="Output the last N lines of the log file, 10 lines are displayed by default",
)
@click.option(
    "--follow",
    "-f",
    "is_follow",
    is_flag=True,
    default=False,
    help="Output appended data as new logs are being generated",
)
def read_logs(lines, is_follow):
    analytics.track(
        "giskard-worker:logs",
        {
            "lines": lines,
            "is_follow": is_follow,
        },
    )
    log_path = get_log_path()

    if not os.path.exists(log_path):
        print(f"Unable to find any logfile!\n{log_path} does not exists")
        exit(-1)

    for line in tail(log_path, lines):
        print(line, end="")

    if is_follow:
        follow_file(log_path)



import pandas as pd
import numpy as np
from dataclasses import dataclass

from giskard import Dataset
from giskard.models import cache as models_cache
from giskard.models.base import BaseModel
from giskard.scanner.decorators import detector
from giskard.scanner.issues import Issue
from giskard.scanner.logger import logger
from ..registry import Detector


@detector(name="data_leakage", tags=["data_leakage", "classification", "regression"])
class DataLeakageDetector(Detector):
    def run(self, model: BaseModel, dataset: Dataset):
        logger.info("DataLeakageDetector: Running")

        # Dataset prediction
        ds_predictions = pd.Series(list(model.predict(dataset).raw), dataset.df.index, dtype=object)

        # Predict on single samples
        sample_idx = dataset.df.sample(min(len(dataset), 100), random_state=23).index.values
        fail_samples = pd.DataFrame(columns=["Whole-dataset prediction", "Single-sample prediction"])

        def slice_single_sample(idx):
            def slice_fn(df):
                return df.loc[[idx]]

            return slice_fn

        for idx, expected_pred in zip(sample_idx, ds_predictions.loc[sample_idx].values):
            row_dataset = dataset.slice(slice_single_sample(idx), row_level=False)
            with models_cache.no_cache():
                row_pred = model.predict(row_dataset).raw[0]

            if not np.isclose(row_pred, expected_pred).all():
                fail_samples.loc[idx, "Whole-dataset prediction"] = expected_pred
                fail_samples.loc[idx, "Single-sample prediction"] = row_pred

            if len(fail_samples) >= 3:
                break

        if not fail_samples.empty:
            return [DataLeakageIssue(model, dataset, level="major", info=DataLeakageInfo(samples=fail_samples))]

        return []


@dataclass
class DataLeakageInfo:
    samples: pd.DataFrame


class DataLeakageIssue(Issue):
    """DataLeakage Issue"""

    group = "Data Leakage"

    @property
    def domain(self) -> str:
        return "Whole dataset"

    @property
    def metric(self) -> str:
        return "Prediction"

    @property
    def deviation(self) -> str:
        return "Model changes output when prediction is run on a single sample"

    @property
    def description(self) -> str:
        return f"We found {len(self.info.samples)} examples for which the model provides a different output depending on whether it is computing on a single data point or on a batch."

    def examples(self, n=3) -> pd.DataFrame:
        return self.info.samples.head(n)

    @property
    def importance(self) -> float:
        return 1

    def generate_tests(self, with_names=False) -> list:
        return []



import pytest
import numpy as np
from unittest.mock import Mock

from giskard.models.base import WrapperModel
from giskard.core.core import ModelMeta
from functools import partial


def _make_model_mock(labels=["First", "Second", "Third"]):
    model = Mock(WrapperModel)
    meta = Mock(ModelMeta)
    meta.classification_labels = labels
    model.meta = meta
    model.model_postprocessing_function = None

    model._convert_to_numpy = partial(WrapperModel._convert_to_numpy, model)
    model._possibly_fix_predictions_shape = partial(WrapperModel._possibly_fix_predictions_shape, model)

    return model


def test_postprocess():
    model = _make_model_mock(["one", "two"])

    model_predict_output = np.array(0.6)
    result = WrapperModel._postprocess(model, model_predict_output)

    assert result.ndim == 2
    assert result.shape[0] == 1
    assert result.shape[1] == 2
    if WrapperModel.is_binary_classification:
        assert result.shape[1] == 2

    model_predict_output = np.array([0.6])
    result = WrapperModel._postprocess(model, model_predict_output)

    assert result.ndim == 2
    assert result.shape[0] == 1
    assert result.shape[1] == 2
    if WrapperModel.is_binary_classification:
        assert result.shape[1] == 2

    model_predict_output = np.array([[0.6, 0.4]])
    result = WrapperModel._postprocess(model, model_predict_output)

    assert result.ndim == 2
    assert result.shape[0] == 1
    assert result.shape[1] == 2
    if WrapperModel.is_binary_classification:
        assert result.shape[1] == 2

    model_predict_output = np.array([[0.6], [0.4]])
    result = WrapperModel._postprocess(model, model_predict_output)

    assert result.ndim == 2
    assert result.shape[0] == 2
    assert result.shape[1] == 2
    if WrapperModel.is_binary_classification:
        assert result.shape[1] == 2


def test_postprocess_fixes_dummy_dimensions_by_squeezing():
    model = _make_model_mock()

    data = np.random.normal(size=(10, 1, 1, 1, 1, 1, 1, 1, 3))
    result = WrapperModel._postprocess(model, data)

    assert result.ndim == 2
    assert result.shape == (10, 3)

    # But should not squeeze first and last dimension
    data = np.random.uniform(size=(1, 3))
    result = WrapperModel._postprocess(model, data)
    assert result.shape == (1, 3)


def test_postprocess_fixes_missing_binary_class():
    model = _make_model_mock(["one", "two"])

    # Should handle 1d data and convert it to (n_entries, 2)
    data = np.random.uniform(size=10)
    result = WrapperModel._postprocess(model, data)

    assert result.shape == (10, 2)
    assert (result[:, 1] == data).all()
    assert np.allclose(result[:, 0], 1 - data)

    # Same for 2d data with a single class
    data = np.random.uniform(size=(10, 1))
    result = WrapperModel._postprocess(model, data)

    assert result.shape == (10, 2)
    assert (result[:, 1] == data[:, 0]).all()
    assert np.allclose(result[:, 0], 1 - data[:, 0])


def test_raises_error_if_classifier_shape_is_not_right():
    # Two classes and 3 outputs: wrong
    model = _make_model_mock(["one", "two"])
    data = np.random.uniform(size=(10, 3))

    with pytest.raises(ValueError):
        _ = WrapperModel._postprocess(model, data)

    # Three classes and 2 outputs: wrong
    model = _make_model_mock(["one", "two", "three"])
    data = np.random.uniform(size=(10, 2))
    with pytest.raises(ValueError):
        _ = WrapperModel._postprocess(model, data)

    # 4 outputs: still wrong
    data = np.random.uniform(size=(10, 4))
    with pytest.raises(ValueError):
        _ = WrapperModel._postprocess(model, data)


def test_should_convert_to_numpy():
    model = _make_model_mock(["one", "two"])
    result = WrapperModel._postprocess(model, [1, 2, 3, 4])
    assert isinstance(result, np.ndarray)



from typing import Optional


class GiskardInstallationError(ImportError):
    functionality: Optional[str]
    flavor: Optional[str]

    def __init__(self, flavor: str = None, functionality: str = None, msg: str = None) -> None:
        self.msg = msg
        if not self.msg:
            self.flavor = flavor or self.flavor
            self.functionality = functionality or self.functionality
            assert self.flavor, "Either giskard package flavor or custom error message should be provided"

            self.msg = (
                f"It seems that you are using Giskard {self.functionality or self.flavor} functionality "
                f"but you are "
                f"missing some required package. Please install Giskard "
                f"with {self.functionality or self.flavor} support "
                f"with `pip install giskard[{self.flavor}]`."
            )



import logging
from abc import ABC
from typing import Callable, Optional, Iterable, Any

import pandas as pd

from giskard.core.core import ModelType
from giskard.models import infer_giskard_cls
from giskard.models.base import CloudpickleBasedModel
from giskard.models.function import PredictionFunctionModel

logger = logging.getLogger(__name__)


class Model(CloudpickleBasedModel, ABC):
    """
    A wrapper class that automatically infers the ML library of the ``model`` argument and provides suitable
    serialization methods (provided by ``save_model`` and ``load_model`` methods).

    Pre-defined serialization and prediction methods cover the ``sklearn``, ``catboost``, ``pytorch``, ``tensorflow``,
    ``huggingface`` and ``langchain`` libraries. If none of these libraries are detected, ``cloudpickle`` is used as the
    default serialization method, and the user is prompted to provide their own prediction method.

    Two ways to wrap your model:

    1. Wrap a prediction function that contains all your data preprocessing steps. It takes as input the raw pandas
       dataframe and returns the probabilities for each classification label or predictions for your regression task.
       Make sure that:

       - ``prediction_function`` encapsulates all the data preprocessing steps (categorical encoding, numerical scaling,
         etc.).
       - ``prediction_function(df[feature_names])`` does not return an error message.

    2. Wrap a model object. This is recommended if your model is not
       serializable by ``cloudpickle`` (e.g., TensorFlow models). This requires:

       - Mandatory: Overriding the ``model_predict`` method, which should take as input the raw pandas dataframe and
         return the probabilities for each classification label (classification), predictions (regression or text generation).
       - Optional: If none of the pre-defined serialization methods apply, ``cloudpickle`` is used as the default
         serialization method. If this fails, the user is asked to override the ``save_model`` and ``load_model``
         methods to provide their own serialization of the model object.

    Args:
        model (Any):
            Could be any function or ML model. The standard model output required for Giskard is:

            * if classification: an array (nxm) of probabilities corresponding to n data entries
              (rows of pandas.DataFrame)
              and m classification_labels. In the case of binary classification, an array of (nx1) probabilities is
              also accepted.
              Make sure that the probability provided is for the second label provided in classification_labels.
            * if regression or text_generation: an array of predictions corresponding to data entries
              (rows of pandas.DataFrame) and outputs.
        name (Optional[str]):
             the name of the model.
        model_type (ModelType):
            The type of the model: regression, classification or text_generation.
        data_preprocessing_function (Optional[Callable[[pd.DataFrame], Any]]):
            A function that takes a pandas.DataFrame as raw input, applies preprocessing and returns any object
            that could be directly fed to clf. You can also choose to include your preprocessing inside clf,
            in which case no need to provide this argument.
        model_postprocessing_function (Optional[Callable[[Any], Any]]):
            A function that takes a clf output as input,
            applies postprocessing and returns an object of the same type and shape as the clf output.
        feature_names (Optional[Iterable[str]]):
            list of feature names matching the column names in the data that correspond to the features which the model
            trained on. By default, feature_names are all the Dataset columns except from target.
        classification_threshold (float):
            represents the classification model threshold, for binary
            classification models.
        classification_labels (Optional[Iterable[str]]):
            that represents the classification labels, if model_type is
            classification. Make sure the labels have the same order as the column output of clf.
        **kwargs: Additional keyword arguments.

    Returns:
        Union[CloudpickleBasedModel, SKLearnModel, HuggingFaceModel,
        CatboostModel, PyTorchModel, TensorFlowModel]: The wrapped Giskard model.
    """

    should_save_model_class = True

    def __new__(
        cls,
        model: Any,
        model_type: ModelType,
        data_preprocessing_function: Callable[[pd.DataFrame], Any] = None,
        model_postprocessing_function: Callable[[Any], Any] = None,
        name: Optional[str] = None,
        feature_names: Optional[Iterable] = None,
        classification_threshold: Optional[float] = 0.5,
        classification_labels: Optional[Iterable] = None,
        **kwargs,
    ):
        """
        Used for dynamical inheritance and returns one of the following class instances:
        ``PredictionFunctionModel``, ``SKLearnModel``, ``CatboostModel``, ``HuggingFaceModel``,
        ``PyTorchModel``, ``TensorFlowModel`` or ``LangchainModel``, depending on the ML library detected in the ``model`` object.
        If the ``model`` object provided does not belong to one of these libraries, an instance of
        ``CloudpickleBasedModel`` is returned in which case:

        1. the default serialization method used will be ``cloudpickle``

        2. you will be asked to provide your own ``model_predict`` method.

        """

        if not model:
            raise ValueError(
                "The 'Model' class cannot be initiated without a `model` argument. "
                "\n`model` can be either a model object (classifier, regressor, etc.) or a prediction function."
                "\nIf a model object is provided, we use a model-tailored serialization method during its upload."
                "\nIf a prediction function is provided, we use cloudpickle to serialize it."
            )
        else:
            giskard_cls = infer_giskard_cls(model)
            # if the Model class is overriden (thus != Model) -> get the methods from the subclass
            # if the Model class is instantiated (thus == Model) -> get the methods from the inferred class
            # if giskard_cls == None -> get the methods from CloudpickleBasedModel
            is_overriden = cls.__name__ != "Model"  # TODO: Improve this
            if is_overriden:
                if not giskard_cls:
                    giskard_cls = CloudpickleBasedModel
                # if save_model and load_model are overriden, replace them, if not, these equalities will be identities.
                possibly_overriden_cls = cls
                possibly_overriden_cls.save_model = giskard_cls.save_model
                possibly_overriden_cls.load_model = giskard_cls.load_model
                possibly_overriden_cls.should_save_model_class = True
            elif giskard_cls:
                input_type = "'prediction_function'" if giskard_cls == PredictionFunctionModel else "'model'"
                logger.info(
                    "Your "
                    + input_type
                    + " is successfully wrapped by Giskard's '"
                    + str(giskard_cls.__name__)
                    + "' wrapper class."
                )
                possibly_overriden_cls = giskard_cls
            else:  # possibly_overriden_cls = CloudpickleBasedModel
                raise NotImplementedError(
                    "We could not infer your model library. You have two options:"
                    "\n- Pass a prediction_function to the Model class "
                    '(we will try to serialize it with "cloudpickle").'
                    "\n- Extend the Model class and override "
                    'the abstract "model_predict" method. Upon upload to the Giskard server, we will try to serialise'
                    'it with "cloudpickle", if that does not work, we will ask you to override the "save_model" and'
                    '"load_model" with your own serialization methods.'
                    "\nWe recommend that you follow our documentation page: "
                    "https://giskard.readthedocs.io/en/latest/getting-started/scan"
                )

            methods = dict(possibly_overriden_cls.__dict__)
            output_cls = type(possibly_overriden_cls.__name__, (giskard_cls,), methods)

            obj = output_cls(
                model=model,
                model_type=model_type,
                data_preprocessing_function=data_preprocessing_function,
                model_postprocessing_function=model_postprocessing_function,
                name=name,
                feature_names=feature_names,
                classification_threshold=classification_threshold,
                classification_labels=classification_labels,
                **kwargs,
            )

            # Important in order for the load method to be executed consistently
            obj.meta.loader_class = possibly_overriden_cls.__name__
            obj.meta.loader_module = possibly_overriden_cls.__module__

            return obj



import inspect
import logging
import posixpath
import tempfile
import uuid
from functools import cached_property
from pathlib import Path
from typing import Dict, Optional, List, Union

import numpy as np
import pandas
import pandas as pd
import yaml
from pandas.api.types import is_list_like
from pandas.api.types import is_numeric_dtype
from xxhash import xxh3_128_hexdigest
from zstandard import ZstdDecompressor

from giskard.client.giskard_client import GiskardClient
from giskard.client.io_utils import save_df, compress
from giskard.client.python_utils import warning
from giskard.core.core import DatasetMeta, SupportedColumnTypes
from giskard.core.validation import configured_validate_arguments
from giskard.ml_worker.testing.registry.slicing_function import SlicingFunction, SlicingFunctionType
from giskard.ml_worker.testing.registry.transformation_function import (
    TransformationFunction,
    TransformationFunctionType,
)
from giskard.settings import settings
from ..metadata.indexing import ColumnMetadataMixin
from ...ml_worker.utils.file_utils import get_file_name

SAMPLE_SIZE = 1000

logger = logging.getLogger(__name__)


class DataProcessor:
    """
    A class for processing tabular data using a pipeline of functions.

    The pipeline consists of slicing functions that extract subsets of the data and transformation functions that modify it.
    Slicing functions should take a pandas DataFrame as input and return a DataFrame, while transformation functions
    should take a DataFrame and return a modified version of it.

    Attributes:
        pipeline (List[Union[SlicingFunction, TransformationFunction]]): a list of functions to be applied to the data,
            in the order in which they were added.

    Methods:
        add_step(processor: Union[SlicingFunction, TransformationFunction]) -> DataProcessor:
            Add a function to the processing pipeline, if it is not already the last step in the pipeline. Return self.

        apply(dataset: Dataset, apply_only_last=False) -> Dataset:
            Apply the processing pipeline to the given dataset. If apply_only_last is True, apply only the last function
            in the pipeline. Return a new Dataset object containing the processed data.

        __repr__() -> str:
            Return a string representation of the DataProcessor object, showing the number of steps in its pipeline.
    """

    pipeline: List[Union[SlicingFunction, TransformationFunction]]

    def __init__(self):
        self.pipeline = []

    @configured_validate_arguments
    def add_step(self, processor: Union[SlicingFunction, TransformationFunction]):
        if not len(self.pipeline) or self.pipeline[-1] != processor:
            self.pipeline.append(processor)
        return self

    def apply(self, dataset: "Dataset", apply_only_last=False):
        ds = dataset.copy()
        is_slicing_only = True

        while len(self.pipeline):
            step = self.pipeline.pop(-1 if apply_only_last else 0)
            is_slicing_only = is_slicing_only and isinstance(step, SlicingFunction)
            df = step.execute(ds) if getattr(step, "needs_dataset", False) else step.execute(ds.df)
            ds = Dataset(
                df=df,
                name=ds.name,
                target=ds.target,
                cat_columns=ds.cat_columns,
                column_types=ds.column_types,
                validation=False,
            )

            if apply_only_last:
                break

        if len(self.pipeline):
            ds.data_processor = self

        # If dataset had metadata, copy it to the new dataset
        if is_slicing_only and hasattr(dataset, "column_meta"):
            ds.load_metadata_from_instance(dataset.column_meta)

        return ds

    def __repr__(self) -> str:
        return f"<DataProcessor ({len(self.pipeline)} steps)>"


class Dataset(ColumnMetadataMixin):
    """
    A class for constructing and processing datasets.

    Attributes:
        df (pandas.DataFrame):
            A `pandas.DataFrame` that contains the raw data (before all the pre-processing steps) and the actual
            ground truth variable (target). `df` can contain more columns than the features of the model, such as the sample_id,
            metadata, etc.
        name (Optional[str]):
            A string representing the name of the dataset (default None).
        target (Optional[str]):
            The column name in df corresponding to the actual target variable (ground truth).
        cat_columns (Optional[List[str]]):
            A list of strings representing the names of categorical columns (default None). If not provided,
            the categorical columns will be automatically inferred.
        column_types (Optional[Dict[str, str]]):
            A dictionary of column names and their types (numeric, category or text) for all columns of df. If not provided,
            the categorical columns will be automatically inferred.
        data_processor (DataProcessor):
            An instance of the `DataProcessor` class used for data processing.
    """

    name: Optional[str]
    target: Optional[str]
    column_types: Dict[str, str]
    df: pd.DataFrame
    id: uuid.UUID
    data_processor: DataProcessor

    @configured_validate_arguments
    def __init__(
        self,
        df: pd.DataFrame,
        name: Optional[str] = None,
        target: Optional[str] = None,
        cat_columns: Optional[List[str]] = None,
        column_types: Optional[Dict[str, str]] = None,
        id: Optional[uuid.UUID] = None,
        validation=True,
    ) -> None:
        """
        Initializes a Dataset object.

        Args:
            df (pd.DataFrame): The input dataset as a pandas DataFrame.
            name (Optional[str]): The name of the dataset.
            target (Optional[str]): The column name in df corresponding to the actual target variable (ground truth).
            cat_columns (Optional[List[str]]): A list of column names that are categorical.
            column_types (Optional[Dict[str, str]]): A dictionary mapping column names to their types.
            id (Optional[uuid.UUID]): A UUID that uniquely identifies this dataset.

        Notes:
            if neither of cat_columns or column_types are provided. We infer heuristically the types of the columns.
            See the _infer_column_types method.
        """
        if id is None:
            self.id = uuid.uuid4()
        else:
            self.id = id
        self.name = name
        self.df = pd.DataFrame(df)
        self.target = target

        if validation:
            from giskard.core.dataset_validation import validate_dtypes, validate_target

            validate_dtypes(self)
            validate_target(self)

        self.column_dtypes = self.extract_column_dtypes(self.df)

        # used in the inference of category columns
        self.category_threshold = round(np.log10(len(self.df))) if len(self.df) >= 100 else 2
        self.column_types = self._infer_column_types(column_types, cat_columns, validation)
        if validation:
            from giskard.core.dataset_validation import validate_column_types

            validate_column_types(self)

        if validation:
            from giskard.core.dataset_validation import validate_column_categorization, validate_numeric_columns

            validate_column_categorization(self)
            validate_numeric_columns(self)

        self.number_of_rows = len(self.df.index)
        self.category_features = {
            column: list(map(lambda x: str(x), self.df[column].dropna().unique()))
            for column, column_type in self.column_types.items()
            if column_type == "category"
        }

        self.data_processor = DataProcessor()

        logger.info("Your 'pandas.DataFrame' is successfully wrapped by Giskard's 'Dataset' wrapper class.")

    def add_slicing_function(self, slicing_function: SlicingFunction):
        """
        Adds a slicing function to the data processor's list of steps.

        Args:
            slicing_function (SlicingFunction): A slicing function to add to the data processor.
        """
        self.data_processor.add_step(slicing_function)
        return self

    def add_transformation_function(self, transformation_function: TransformationFunction):
        """
        Add a transformation function to the data processor's list of steps.

        Args:
            transformation_function (TransformationFunction): A transformation function to add to the data processor.
        """
        self.data_processor.add_step(transformation_function)
        return self

    @cached_property
    def row_hashes(self):
        return pandas.Series(
            map(
                lambda row: xxh3_128_hexdigest(f"{', '.join(map(lambda x: repr(x), row))}".encode("utf-8")),
                self.df.values,
            ),
            index=self.df.index,
        )

    @configured_validate_arguments
    def slice(
        self,
        slicing_function: Union[SlicingFunction, SlicingFunctionType],
        row_level: bool = True,
        cell_level=False,
        column_name: Optional[str] = None,
    ):
        """
        Slice the dataset using the specified `slicing_function`.

        Args:
            slicing_function (Union[SlicingFunction, SlicingFunctionType]): A slicing function to apply.
                If `slicing_function` is a callable, it will be wrapped in a `SlicingFunction` object
                with `row_level` and `cell_level` as its arguments. The `SlicingFunction` object will be
                used to slice the DataFrame. If `slicing_function` is a `SlicingFunction` object, it
                will be used directly to slice the DataFrame.
            row_level (bool): Whether the `slicing_function` should be applied to the rows (True) or
                the whole dataframe (False). Defaults to True.
            cell_level (bool): Whether the `slicing_function` should be applied to the cells (True) or
                the whole dataframe (False). Defaults to False.

        Returns:
            Dataset:
                The sliced dataset as a `Dataset` object.

        Notes:
            Raises TypeError: If `slicing_function` is not a callable or a `SlicingFunction` object.
        """
        if inspect.isfunction(slicing_function):
            slicing_function = SlicingFunction(slicing_function, row_level=row_level, cell_level=cell_level)

        if slicing_function.cell_level and column_name is not None:
            slicing_function = slicing_function(
                column_name=column_name,
                **{key: value for key, value in slicing_function.params.items() if key != "column_name"},
            )

        return self.data_processor.add_step(slicing_function).apply(self, apply_only_last=True)

    @configured_validate_arguments
    def transform(
        self,
        transformation_function: Union[TransformationFunction, TransformationFunctionType],
        row_level: bool = True,
        cell_level=False,
        column_name: Optional[str] = None,
    ):
        """
        Transform the data in the current Dataset by applying a transformation function.

        Args:
            transformation_function (Union[TransformationFunction, TransformationFunctionType]):
                A transformation function to apply. If `transformation_function` is a callable, it will
                be wrapped in a `TransformationFunction` object with `row_level` and `cell_level` as its
                arguments. If `transformation_function` is a `TransformationFunction` object, it will be used
                directly to transform the DataFrame.
            row_level (bool): Whether the `transformation_function` should be applied to the rows (True) or
                the whole dataframe (False). Defaults to True.
            cell_level (bool): Whether the `slicing_function` should be applied to the cells (True) or
                the whole dataframe (False). Defaults to False.

        Returns:
            Dataset: A new Dataset object containing the transformed data.

        Notes:
            Raises TypeError: If `transformation_function` is not a callable or a `TransformationFunction` object.
        """

        if inspect.isfunction(transformation_function):
            transformation_function = TransformationFunction(
                transformation_function, row_level=row_level, cell_level=cell_level
            )

        if transformation_function.cell_level and column_name is not None:
            transformation_function = transformation_function(
                column_name=column_name,
                **{key: value for key, value in transformation_function.params.items() if key != "column_name"},
            )

        assert (
            not transformation_function.cell_level or "column_name" in transformation_function.params
        ), "column_name should be provided for TransformationFunction at cell level"
        return self.data_processor.add_step(transformation_function).apply(self, apply_only_last=True)

    def process(self):
        """
        Process the dataset by applying all the transformation and slicing functions in the defined order.

        Returns:
            The processed dataset after applying all the transformation and slicing functions.
        """
        return self.data_processor.apply(self)

    def _infer_column_types(
        self, column_types: Optional[Dict[str, str]], cat_columns: Optional[List[str]], validation: bool = True
    ):
        """
        This function infers the column types of a given DataFrame based on the number of unique values and column data types. It takes into account the provided column types and categorical columns. The inferred types can be 'text', 'numeric', or 'category'. The function also applies a logarithmic rule to determine the category threshold.

        Here's a summary of the function's logic:

        1. If no column types are provided, initialize an empty dictionary.
        2. Determine the columns in the DataFrame, excluding the target column if it exists.
        3. If categorical columns are specified, prioritize them over the provided column types and mark them as 'category'.
        4. Check for any unknown columns in the provided column types and remove them from the dictionary.
        5. If there are no missing columns, remove the target column (if present) from the column types dictionary.
        6. Calculate the number of unique values in each missing column.
        7. For each missing column:

           - If the number of unique values is less than or equal to the category threshold, categorize it as 'category'.
           - Otherwise, attempt to convert the column to numeric using `pd.to_numeric` and categorize it as 'numeric'.
           - If the column does not have the expected numeric data type and validation is enabled, issue a warning message.
           - If conversion to numeric raises a ValueError, categorize the column as 'text'.
        8. Return the column types dictionary.

        The logarithmic rule is used to calculate the category threshold. The formula is: `category_threshold = round(np.log10(len(self.df))) if len(self.df) >= 100 else 2`. This means that if the length of the DataFrame is greater than or equal to 100, the category threshold is set to the rounded value of the base-10 logarithm of the DataFrame length. Otherwise, the category threshold is set to 2. The logarithmic rule helps in dynamically adjusting the category threshold based on the size of the DataFrame.

        Returns:
           dict: A dictionary that maps column names to their inferred types, one of 'text', 'numeric', or 'category'.
        """
        if not column_types:
            column_types = {}
        df_columns = set([col for col in self.columns if col != self.target]) if self.target else set(self.columns)

        # priority of cat_columns over column_types (for categorical columns)
        if cat_columns:
            if not set(cat_columns).issubset(df_columns):
                raise ValueError(
                    "The provided 'cat_columns' are not all part of your dataset 'columns'. "
                    "Please make sure that `cat_columns` refers to existing columns in your dataset."
                )
            for cat_col in cat_columns:
                if cat_col != self.target:
                    column_types[cat_col] = SupportedColumnTypes.CATEGORY.value

        given_columns = set(column_types.keys())
        unknown_columns = given_columns - df_columns
        missing_columns = df_columns - given_columns

        if unknown_columns:
            warning(
                f"The provided keys {list(unknown_columns)} in 'column_types' are not part of your dataset "
                "'columns'. Please make sure that the column names in `column_types` refers to existing "
                "columns in your dataset."
            )
            [column_types.pop(i) for i in unknown_columns]

        if not missing_columns:
            column_types.pop(self.target, None)  # no need for target type
            return column_types

        nuniques = self.df.nunique()
        for col in missing_columns:
            if col == self.target:
                continue
            if nuniques[col] <= self.category_threshold:
                column_types[col] = SupportedColumnTypes.CATEGORY.value
                continue
            # inference of text and numeric columns
            try:
                pd.to_numeric(self.df[col])
                column_types[col] = SupportedColumnTypes.NUMERIC.value
                if not is_numeric_dtype(self.df[col]) and validation:
                    warning(
                        f"The column {col} is declared as numeric but has '{str(self.df[col].dtype)}' as data type. "
                        "To avoid potential future issues, make sure to cast this column to the correct data type."
                    )

            except ValueError:
                column_types[col] = SupportedColumnTypes.TEXT.value

        return column_types

    @staticmethod
    def extract_column_dtypes(df):
        """
        Extracts the column data types from a pandas DataFrame.

        Args:
            df (pandas.DataFrame): The input DataFrame.

        Returns:
            dict: A dictionary where the keys are the column names and the values are the corresponding data types as strings.
        """
        return df.dtypes.apply(lambda x: x.name).to_dict()

    def upload(self, client: GiskardClient, project_key: str):
        """
        Uploads the dataset to the specified Giskard project.

        Args:
            client: A GiskardClient instance for connecting to the Giskard API.
            project_key (str): The key of the project to upload the dataset to.

        Returns:
            str: The ID of the uploaded dataset.
        """
        dataset_id = str(self.id)

        with tempfile.TemporaryDirectory(prefix="giskard-dataset-") as local_path:
            original_size_bytes, compressed_size_bytes = self.save(Path(local_path), dataset_id)
            client.log_artifacts(local_path, posixpath.join(project_key, "datasets", dataset_id))
            client.save_dataset_meta(
                project_key,
                dataset_id,
                self.meta,
                original_size_bytes=original_size_bytes,
                compressed_size_bytes=compressed_size_bytes,
            )
        return dataset_id

    @property
    def meta(self):
        return DatasetMeta(
            name=self.name,
            target=self.target,
            column_types=self.column_types,
            column_dtypes=self.column_dtypes,
            number_of_rows=self.number_of_rows,
            category_features=self.category_features,
        )

    @staticmethod
    def cast_column_to_dtypes(df, column_dtypes):
        current_types = df.dtypes.apply(lambda x: x.name).to_dict()
        logger.info(f"Casting dataframe columns from {current_types} to {column_dtypes}")
        if column_dtypes:
            try:
                df = df.astype(column_dtypes, errors="ignore")
            except Exception as e:
                raise ValueError("Failed to apply column types to dataset") from e
        return df

    @classmethod
    def load(cls, local_path: str):
        with open(local_path, "rb") as ds_stream:
            return pd.read_csv(
                ZstdDecompressor().stream_reader(ds_stream),
                keep_default_na=False,
                na_values=["_GSK_NA_"],
            )

    @classmethod
    def download(cls, client: GiskardClient, project_key, dataset_id, sample: bool = False):
        """
        Downloads a dataset from a Giskard project and returns a Dataset object.
        If the client is None, then the function assumes that it is running in an internal worker and looks for the dataset locally.

        Args:
            client (GiskardClient):
                The GiskardClient instance to use for downloading the dataset.
                If None, the function looks for the dataset locally.
            project_key (str): The key of the Giskard project that the dataset belongs to.
            dataset_id (str): The ID of the dataset to download.
            sample (bool): Only open a sample of 1000 rows if True

        Returns:
            Dataset: A Dataset object that represents the downloaded dataset.
        """
        local_dir = settings.home_dir / settings.cache_dir / project_key / "datasets" / dataset_id

        if client is None:
            # internal worker case, no token based http client
            assert local_dir.exists(), f"Cannot find existing dataset {project_key}.{dataset_id}"
            with open(Path(local_dir) / "giskard-dataset-meta.yaml") as f:
                saved_meta = yaml.load(f, Loader=yaml.Loader)
                meta = DatasetMeta(
                    name=saved_meta["name"],
                    target=saved_meta["target"],
                    column_types=saved_meta["column_types"],
                    column_dtypes=saved_meta["column_dtypes"],
                    number_of_rows=saved_meta["number_of_rows"],
                    category_features=saved_meta["category_features"],
                )
        else:
            client.load_artifact(local_dir, posixpath.join(project_key, "datasets", dataset_id))
            meta: DatasetMeta = client.load_dataset_meta(project_key, dataset_id)

        df = cls.load(local_dir / get_file_name("data", "csv.zst", sample))
        df = cls.cast_column_to_dtypes(df, meta.column_dtypes)
        return cls(
            df=df,
            name=meta.name,
            target=meta.target,
            column_types=meta.column_types,
            id=uuid.uuid4() if sample else uuid.UUID(dataset_id),
        )

    @staticmethod
    def _cat_columns(meta):
        return (
            [fname for (fname, ftype) in meta.column_types.items() if ftype == SupportedColumnTypes.CATEGORY]
            if meta.column_types
            else None
        )

    @property
    def cat_columns(self):
        return self._cat_columns(self.meta)

    def save(self, local_path: Path, dataset_id):
        with open(local_path / "data.csv.zst", "wb") as f, open(local_path / "data.sample.csv.zst", "wb") as f_sample:
            uncompressed_bytes = save_df(self.df)
            compressed_bytes = compress(uncompressed_bytes)
            f.write(compressed_bytes)
            original_size_bytes, compressed_size_bytes = len(uncompressed_bytes), len(compressed_bytes)

            uncompressed_bytes = save_df(self.df.sample(min(SAMPLE_SIZE, len(self.df.index))))
            compressed_bytes = compress(uncompressed_bytes)
            f_sample.write(compressed_bytes)

            with open(Path(local_path) / "giskard-dataset-meta.yaml", "w") as meta_f:
                yaml.dump(
                    {
                        "id": dataset_id,
                        "name": self.meta.name,
                        "target": self.meta.target,
                        "column_types": self.meta.column_types,
                        "column_dtypes": self.meta.column_dtypes,
                        "original_size_bytes": original_size_bytes,
                        "compressed_size_bytes": compressed_size_bytes,
                        "number_of_rows": self.meta.number_of_rows,
                        "category_features": self.meta.category_features,
                    },
                    meta_f,
                    default_flow_style=False,
                )
            return original_size_bytes, compressed_size_bytes

    @property
    def columns(self):
        return self.df.columns

    def __len__(self):
        return len(self.df)

    def select_columns(self, columns=None, col_type=None):
        columns = _cast_to_list_like(columns) if columns is not None else None
        col_type = _cast_to_list_like(col_type) if col_type is not None else None

        df = self.df.copy()

        if columns is None and col_type is None:
            # TODO: should probably copy
            return self

        # Filter by columns
        if columns is not None:
            df = df.loc[:, columns]

        # Filter by type
        if col_type is not None:
            if not is_list_like(col_type):
                col_type = [col_type]

            columns = [col for col in df.columns if self.column_types[col] in col_type]
            df = df.loc[:, columns]

        return Dataset(
            df=df,
            target=self.target if self.target in df.columns else None,
            column_types={key: val for key, val in self.column_types.items() if key in df.columns},
            validation=False,
        )

    def copy(self):
        dataset = Dataset(
            df=self.df.copy(),
            target=self.target,
            column_types=self.column_types.copy(),
            validation=False,
        )

        if hasattr(self, "column_meta"):
            dataset.load_metadata_from_instance(self.column_meta)

        return dataset


def _cast_to_list_like(object):
    return object if is_list_like(object) else (object,)



import numpy as np
import pandas as pd
import pytest

import giskard.testing.tests.metamorphic as metamorphic
from giskard.ml_worker.testing.registry.transformation_function import transformation_function
from giskard.ml_worker.testing.stat_utils import equivalence_t_test, paired_t_test
from giskard.ml_worker.testing.stat_utils import equivalence_wilcoxon, paired_wilcoxon
from giskard.ml_worker.testing.utils import Direction


def test_metamorphic_invariance_no_change(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_invariance(
        model=german_credit_model, dataset=german_credit_test_data, transformation_function=perturbation, threshold=0.1
    ).execute()

    assert results.actual_slices_size[0] == 1000
    assert results.passed


def test_metamorphic_invariance_with_non_linear_index(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    dataset = german_credit_test_data.slice(lambda df: df.sample(100), row_level=False)

    results = metamorphic.test_metamorphic_invariance(
        model=german_credit_model, dataset=dataset, transformation_function=perturbation, threshold=0.1
    ).execute()

    assert results.passed


def _test_metamorphic_invariance_male_female(german_credit_test_data, german_credit_model, threshold):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_invariance(
        model=german_credit_model,
        dataset=german_credit_test_data,
        transformation_function=perturbation,
        threshold=threshold,
    ).execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert round(results.metric, 2) == 0.97
    return results.passed


def test_metamorphic_invariance_male_female(german_credit_test_data, german_credit_model):
    assert _test_metamorphic_invariance_male_female(german_credit_test_data, german_credit_model, 0.5)
    assert not _test_metamorphic_invariance_male_female(german_credit_test_data, german_credit_model, 1)


def test_metamorphic_invariance_2_perturbations(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.duration_in_month = x.duration_in_month * 2
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_invariance(
        model=german_credit_model, dataset=german_credit_test_data, transformation_function=perturbation, threshold=0.1
    ).execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert round(results.metric, 2) == 0.86


def test_metamorphic_invariance_some_rows(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_invariance(
        model=german_credit_model, dataset=german_credit_test_data, transformation_function=perturbation, threshold=0.1
    ).execute()

    assert round(results.metric, 2) == 0.97


def test_metamorphic_invariance_regression(diabetes_dataset_with_target, linear_regression_diabetes):
    sex_values = list(diabetes_dataset_with_target.df.sex.unique())
    assert len(sex_values) == 2

    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = sex_values[1] if x.sex == sex_values[0] else x.sex
        return x

    results = metamorphic.test_metamorphic_invariance(
        model=linear_regression_diabetes,
        dataset=diabetes_dataset_with_target,
        transformation_function=perturbation,
        threshold=0.1,
        output_sensitivity=0.1,
    ).execute()

    assert results.actual_slices_size[0] == len(diabetes_dataset_with_target)
    assert round(results.metric, 2) == 0.11


@pytest.mark.parametrize(
    "loc_pop, scale_pop, loc_perturb, scale_perturb, direction, window_size, critical_quantile",
    [
        (0.5, 0.5, 0, 1e-2, Direction.Invariant, 0.2, 0.05),
        (0.5, 0.5, 0.1, 0.1, Direction.Increasing, float("nan"), 0.05),
        (0.5, 0.5, -0.1, 0.1, Direction.Decreasing, float("nan"), 0.05),
    ],
)
def test_metamorphic_compare_t_test(
    loc_pop, scale_pop, loc_perturb, scale_perturb, direction, window_size, critical_quantile
):
    population = np.random.normal(loc_pop, scale_pop, size=100)
    perturbation = np.random.normal(loc_perturb, scale_perturb, size=100)
    print(perturbation)

    result_inc, p_value_inc = paired_t_test(
        population, population + perturbation, alternative="less", critical_quantile=critical_quantile
    )
    result_dec, p_value_dec = paired_t_test(
        population, population + perturbation, alternative="greater", critical_quantile=critical_quantile
    )
    result_inv, p_value_inv = equivalence_t_test(
        population, population + perturbation, window_size=window_size, critical_quantile=critical_quantile
    )

    dict_mapping = {
        Direction.Invariant: (result_inv, p_value_inv),
        Direction.Decreasing: (result_dec, p_value_dec),
        Direction.Increasing: (result_inc, p_value_inc),
    }

    print(f"Direction: {direction.name} and p_value; {dict_mapping[direction][1]}")
    assert dict_mapping[direction][1] < critical_quantile


@pytest.mark.parametrize(
    "loc_pop, scale_pop, loc_perturb, scale_perturb, direction, window_size, critical_quantile",
    [
        (0.5, 0.5, 0, 1e-2, Direction.Invariant, 0.2, 0.05),
        (0.5, 0.5, 0.1, 0.1, Direction.Increasing, float("nan"), 0.05),
        (0.5, 0.5, -0.1, 0.1, Direction.Decreasing, float("nan"), 0.05),
    ],
)
def test_metamorphic_compare_wilcoxon(
    loc_pop, scale_pop, loc_perturb, scale_perturb, direction, window_size, critical_quantile
):
    population = np.random.normal(loc_pop, scale_pop, size=100)
    perturbation = np.random.normal(loc_perturb, scale_perturb, size=100)

    result_inc, p_value_inc = paired_wilcoxon(
        population, population + perturbation, alternative="less", critical_quantile=critical_quantile
    )
    result_dec, p_value_dec = paired_wilcoxon(
        population, population + perturbation, alternative="greater", critical_quantile=critical_quantile
    )
    result_inv, p_value_inv = equivalence_wilcoxon(
        population, population + perturbation, window_size=window_size, critical_quantile=critical_quantile
    )

    dict_mapping = {
        Direction.Invariant: (result_inv, p_value_inv),
        Direction.Decreasing: (result_dec, p_value_dec),
        Direction.Increasing: (result_inc, p_value_inc),
    }

    print(f"Direction: {direction.name} and p_value; {dict_mapping[direction][1]}")
    assert dict_mapping[direction][1] < critical_quantile


def test_metamorphic_invariance_t_test(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_invariance_t_test(
        model=german_credit_model,
        dataset=german_credit_test_data,
        transformation_function=perturbation,
        window_size=0.2,
        critical_quantile=0.05,
    ).execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert results.passed, f"metric = {results.metric}"


def test_metamorphic_invariance_t_test_nopert(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_invariance_t_test(
        model=german_credit_model,
        dataset=german_credit_test_data,
        transformation_function=perturbation,
        window_size=0.2,
        critical_quantile=0.05,
    ).execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert results.passed, f"metric = {results.metric}"


def test_metamorphic_invariance_wilcoxon(german_credit_test_data, german_credit_model):
    @transformation_function
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_invariance_wilcoxon(
        model=german_credit_model,
        dataset=german_credit_test_data,
        transformation_function=perturbation,
        window_size=0.2,
        critical_quantile=0.05,
    ).execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert results.passed, f"metric = {results.metric}"


def test_metamorphic_invariance_wilcoxon_nopert(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_invariance_wilcoxon(
        model=german_credit_model,
        dataset=german_credit_test_data,
        transformation_function=perturbation,
        window_size=0.2,
        critical_quantile=0.05,
    ).execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert results.passed, f"metric = {results.metric}"



__all__ = [
    "test_drift_psi",
    "test_drift_chi_square",
    "test_drift_ks",
    "test_drift_earth_movers_distance",
    "test_drift_prediction_psi",
    "test_drift_prediction_chi_square",
    "test_drift_prediction_ks",
    "test_drift_prediction_earth_movers_distance",
    "test_right_label",
    "test_output_in_range",
    "test_disparate_impact",
    "test_mae",
    "test_rmse",
    "test_recall",
    "test_auc",
    "test_accuracy",
    "test_precision",
    "test_f1",
    "test_r2",
    "test_diff_recall",
    "test_diff_accuracy",
    "test_diff_precision",
    "test_diff_rmse",
    "test_diff_f1",
    "test_metamorphic_invariance",
    "test_metamorphic_increasing",
    "test_metamorphic_decreasing",
    "test_metamorphic_decreasing_t_test",
    "test_metamorphic_increasing_t_test",
    "test_metamorphic_invariance_t_test",
    "test_metamorphic_increasing_wilcoxon",
    "test_metamorphic_decreasing_wilcoxon",
    "test_metamorphic_invariance_wilcoxon",
]

from giskard.testing.tests.drift import (
    test_drift_psi,
    test_drift_chi_square,
    test_drift_ks,
    test_drift_earth_movers_distance,
    test_drift_prediction_psi,
    test_drift_prediction_chi_square,
    test_drift_prediction_ks,
    test_drift_prediction_earth_movers_distance,
)
from giskard.testing.tests.metamorphic import (
    test_metamorphic_invariance,
    test_metamorphic_increasing,
    test_metamorphic_decreasing,
    test_metamorphic_decreasing_t_test,
    test_metamorphic_increasing_t_test,
    test_metamorphic_invariance_t_test,
    test_metamorphic_decreasing_wilcoxon,
    test_metamorphic_increasing_wilcoxon,
    test_metamorphic_invariance_wilcoxon,
)
from giskard.testing.tests.performance import (
    test_auc,
    test_f1,
    test_accuracy,
    test_precision,
    test_recall,
    test_rmse,
    test_mae,
    test_r2,
    test_diff_accuracy,
    test_diff_f1,
    test_diff_precision,
    test_diff_recall,
    test_diff_rmse,
)
from giskard.testing.tests.statistic import (
    test_right_label,
    test_output_in_range,
    test_disparate_impact,
)



import numpy as np
import pandas as pd
from typing import Optional

from ...ml_worker.testing.registry.decorators import test

from ...ml_worker.testing.test_result import TestResult

from ...ml_worker.testing.registry.slicing_function import SlicingFunction
from ...datasets.base import Dataset
from ...models.base import BaseModel


def _calculate_overconfidence_score(model: BaseModel, dataset: Dataset) -> pd.Series:
    true_target = dataset.df.loc[:, dataset.target].values
    pred = model.predict(dataset)
    label2id = {label: n for n, label in enumerate(model.meta.classification_labels)}

    # Empirical cost associated to overconfidence, i.e. the difference between
    # the probability assigned to the predicted label and the correct label.
    p_max = pred.probabilities
    p_true_label = np.array([pred.raw[n, label2id[label]] for n, label in enumerate(true_target)])

    overconfidence_score = p_max - p_true_label
    return pd.Series(overconfidence_score, index=dataset.df.index)


def _default_overconfidence_threshold(model: BaseModel) -> float:
    n = len(model.meta.classification_labels)
    return 1 / (3e-1 * (n - 2) + 2 - 1e-3 * (n - 2) ** 2)


@test(name="Overconfidence Rate", tags=["classification"])
def test_overconfidence_rate(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: Optional[float] = 0.10,
    p_threshold: Optional[float] = None,
):
    """Tests that the rate of overconfident predictions is below a threshold.

    Overconfident predictions are defined as predictions where the model
    assigned a large probability to the wrong label. We quantify this as the
    difference between the largest probability assigned to a label and the
    probability assigned to the correct label (this will be 0 if the model
    made the correct prediction). If this is larger than a threshold
    (`p_threshold`, typically determined automatically depending on the number
    of classes), then the prediction is considered overconfident.
    We then calculate the rate of overconfident predictions as the number of
    overconfident samples divided by the total number of wrongly predicted
    samples, and check that it is below a user-specified threshold.

    Arguments:
        model(BaseModel): The model to test.
        dataset(Dataset): The dataset to test the model on.
        slicing_function(SlicingFunction, optional): An optional slicing
            function used to slice the dataset before testing. If not provided,
            the whole dataset will be considered in calculating the
            overconfidence rate.
        threshold(float, optional): The threshold for overconfident prediction
            rate, i.e. the max ratio of overconfident samples over number of
            wrongly predicted samples. Default is 0.10 (10%).
        p_threshold(float, optional): The threshold for the difference between
            the probability assigned to the wrong label and the correct label
            over which a prediction is considered overconfident. If not
            provided, it will be determined automatically depending on the
            number of classes.
    """
    if not model.is_classification:
        raise ValueError("This test is only applicable to classification models.")

    if slicing_function is not None:
        dataset = dataset.slice(slicing_function)

    overconfidence_score = _calculate_overconfidence_score(model, dataset)

    if p_threshold is None:
        p_threshold = _default_overconfidence_threshold(model)

    rate = (overconfidence_score[overconfidence_score > 0].dropna() > p_threshold).mean()
    passed = rate < threshold

    return TestResult(
        passed=bool(passed),
        metric=rate,
    )


def _calculate_underconfidence_score(model: BaseModel, dataset: Dataset) -> pd.Series:
    # Empirical cost associated to underconfidence: difference between the two
    # most probable classes.
    ps = model.predict(dataset).raw

    # Relative difference
    ps_2 = -np.partition(-ps, 1, axis=-1)[:, :2]
    score_values = ps_2.min(axis=-1) / ps_2.max(axis=-1)

    return pd.Series(score_values, index=dataset.df.index)


@test(name="Underconfidence Rate", tags=["classification"])
def test_underconfidence_rate(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: Optional[float] = 0.10,
    p_threshold: float = 0.90,
):
    """Tests that the rate of underconfident predictions is below a threshold.

    Underconfident predictions are defined as predictions where the two most
    probable labels have very similar probabilities. In this case, slight
    changes can make the model flip its predicted label. By default, we mark a
    prediction as underconfident when the second most probable prediction has a
    probability which is only less than 10% smaller than the predicted label
    (`p_threshold=0.90`).
    We then calculate the rate of underconfident predictions as the number of
    underconfident samples divided by the total number of samples, and check
    that it is below the user-specified threshold.


    Arguments:
        model(BaseModel): The model to test.
        dataset(Dataset): The dataset to test the model on.
        slicing_function(SlicingFunction, optional): An optional slicing
            function used to slice the dataset before testing. If not provided,
            the whole dataset will be considered in calculating the
            underconfidence rate.
        threshold(float, optional): The threshold for underconfident prediction
            rate. Default is 0.10 (10%).
        p_threshold(float, optional): The threshold for the relative value of
            the second most-probable prediction and the max probability. If
            greater that this value, the prediction is considered
            underconfident. Default is 0.90, i.e. when the second most probable
            prediction is 90% or more with respect to the highest probability,
            the sample prediction is considered underconfident.
    """
    if not model.is_classification:
        raise ValueError("This test is only applicable to classification models.")

    if slicing_function is not None:
        dataset = dataset.slice(slicing_function)

    underconfidence_score = _calculate_underconfidence_score(model, dataset)

    rate = (underconfidence_score.dropna() > p_threshold).mean()
    passed = rate < threshold

    return TestResult(
        passed=bool(passed),
        metric=rate,
    )



from giskard.models.sklearn import SKLearnModel
import mlflow


class CatboostModel(SKLearnModel):
    """
    A subclass of the SKLearnModel class for using the Catboost model.

    The CatboostModel class is a wrapper around the Catboost machine learning library.

    Inherits all attributes and methods from the SKLearnModel class.

    Attributes:
        _feature_names_attr (str): A string attribute indicating the name of the feature names attribute for the Catboost model.

    """

    _feature_names_attr = "feature_names_"

    def save_model(self, local_path, mlflow_meta: mlflow.models.Model):
        mlflow.catboost.save_model(self.model, path=local_path, mlflow_model=mlflow_meta)



import tempfile
from pathlib import Path

import numpy as np
import pytest

from giskard import Model
from giskard.core.core import SupportedModelTypes


def test_catboost(german_credit_test_data, german_credit_catboost):
    res = german_credit_catboost.predict(german_credit_test_data)
    assert len(res.prediction) == len(german_credit_test_data.df)
    assert (res.probabilities >= 0).all() and (res.probabilities <= 1).all()


@pytest.mark.skip(reason="GSK-230 enable test once model upload API is changed from functional to class based")
def test_catboost_changed_column_order(german_credit_test_data, german_credit_catboost):
    original_predictions = german_credit_catboost.predict(german_credit_test_data).probabilities

    # change column order
    df = german_credit_test_data.df
    german_credit_test_data.df = df.reindex(df.columns[::-1], axis=1)

    # reset feature names to test the behaviour when they're not provided
    german_credit_catboost.feature_names = None

    res = german_credit_catboost.predict(german_credit_test_data)
    assert len(res.prediction) == len(german_credit_test_data.df)
    assert (res.probabilities >= 0).all() and (res.probabilities <= 1).all()
    assert np.array_equal(
        german_credit_catboost.predict(german_credit_test_data).probabilities,
        original_predictions,
    ), "Predictions are not the same after changing features order"


def test_prediction_cache_loaded_model(linear_regression_diabetes, linear_regression_diabetes_raw, diabetes_dataset):
    nb_of_prediction_calls = [0]

    def prediction_fn(df):
        nb_of_prediction_calls[0] += 1
        return np.ones(df.shape[0])

    with tempfile.TemporaryDirectory(prefix="cache_dir-") as cache_dir, tempfile.TemporaryDirectory(
        prefix="model_dir-"
    ) as model_dir:
        model = Model(
            prediction_fn,
            model_type=SupportedModelTypes.REGRESSION,
            feature_names=linear_regression_diabetes.meta.feature_names,
            prediction_cache_dir=Path(cache_dir),
        )

        predictions = model.predict(diabetes_dataset)
        model.save(model_dir)

        loaded_model = Model.load(model_dir, prediction_cache_dir=Path(cache_dir))
        second_predictions = loaded_model.predict(diabetes_dataset)

        assert np.all(np.equal(predictions.raw, second_predictions.raw))
        assert nb_of_prediction_calls[0] == 1



import asyncio
import base64
import logging
import sys
from asyncio import StreamReader, StreamWriter, IncompleteReadError
from os import environ
from random import random
from tenacity import retry, wait_exponential, stop_after_attempt, after_log
from typing import Optional
from urllib.parse import urlparse

from giskard.client.giskard_client import GiskardClient
from giskard.ml_worker.bridge.data_encryptor import DataEncryptor
from giskard.ml_worker.bridge.error import ConnectionLost
from giskard.ml_worker.bridge.service_messages import CREATE_CLIENT_CHANNEL, START_INNER_SERVER
from giskard.ml_worker.utils.network import readable_hex
from giskard.utils.analytics_collector import anonymize, analytics

CHANNEL_ID_LENGTH = 8

logger = logging.getLogger(__name__)

# Some python versions (3.10 for examples) may garbage collect readers that are in use
# it leads to GeneratorExit errors in coroutines and failed GRPC calls
# this set is meant to keep readers references so that GC doesn't collect them
readers = set()
writers = set()


class MLWorkerBridge:
    key_id: str
    encryptor: DataEncryptor
    remote_host: str
    remote_port: int

    def __init__(
        self,
        local_address: str,
        client: GiskardClient,
    ) -> None:
        self.stopping = False
        self.local_address = local_address

        self.client = client

        self.service_channel_reader: Optional[StreamReader] = None
        self.service_channel_writer: Optional[StreamWriter] = None

    async def stop(self):
        logger.info(f"Stopping the bridge with {self.local_address}")
        self.stopping = True
        await self.close_service_channel()
        for w in writers:
            w.close()

    @retry(wait=wait_exponential(min=0.1, max=5, multiplier=0.1))
    async def start(self):
        try:
            await self.connect_to_remote_host()
            await self.send_start_inner_server_message()
            await self.listen_remote_server_service_socket()

        except Exception as e:
            await self.close_service_channel()
            logger.error(f"Failed to connect to a remote host: {self.remote_host}:{self.remote_port} : {str(e)}")
            analytics.track(
                "Start ML Worker Bridge error",
                {"host": anonymize(self.remote_host), "port": anonymize(self.remote_port), "error": str(e)},
                force=True,
            )
            raise e

    async def close_service_channel(self):
        if self.service_channel_writer:
            self.service_channel_writer.close()
            await self.service_channel_writer.wait_closed()
            writers.remove(self.service_channel_writer)
        if self.service_channel_reader:
            readers.remove(self.service_channel_reader)

    def fetch_connection_details(self):
        info = self.client.get_server_info()
        self.remote_host = (
            environ.get("GSK_EXTERNAL_ML_WORKER_HOST")
            or info.get("externalMlWorkerEntrypointHost")
            or urlparse(self.client.host_url).hostname
        )

        self.remote_port = (
            environ.get("GSK_EXTERNAL_ML_WORKER_PORT") or info.get("externalMlWorkerEntrypointPort") or 40051
        )

        encryption_key = base64.b64decode(info["encryptionKey"])

        self.key_id = info["keyId"]
        self.encryptor = DataEncryptor(encryption_key)

    async def connect_to_remote_host(self):
        self.fetch_connection_details()

        logger.info(f"Connecting to {self.remote_host}:{self.remote_port}")
        self.service_channel_reader, self.service_channel_writer = await asyncio.open_connection(
            self.remote_host, self.remote_port
        )
        readers.add(self.service_channel_reader)
        writers.add(self.service_channel_writer)
        logger.info(f"Connected to Giskard server {self.remote_host}:{self.remote_port}")

    async def send_start_inner_server_message(self):
        self.service_channel_writer.write(
            self.encryptor.encrypt(START_INNER_SERVER, additional_data=self.key_id.encode())
        )
        await self.service_channel_writer.drain()

    async def listen_remote_server_service_socket(self):
        try:
            logger.debug("Created remote server listener task")
            while True:
                logger.debug("waiting for a service command")
                length = int.from_bytes(await self.service_channel_reader.readexactly(4), "big")
                encrypted_data = await self.service_channel_reader.readexactly(length)
                if len(encrypted_data):
                    data = self.encryptor.decrypt(encrypted_data)
                    client = data[:CHANNEL_ID_LENGTH]
                    command = data[CHANNEL_ID_LENGTH:]
                    logger.debug(f"service command received: {client}: {command}")
                    await self.handle_server_command(client, command)
                else:
                    raise ConnectionLost()
        except (IncompleteReadError, ConnectionLost, ConnectionResetError) as e:
            if self.stopping:
                return
            logger.info("Lost connection to Giskard server, retrying...")
            await self.stop()
            await asyncio.sleep(1 + random() * 2)
            await self.start()
            if isinstance(e, IncompleteReadError) and not self.stopping:
                logger.exception(e)
        except BaseException as e:  # NOSONAR
            logger.exception(e)

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        after=after_log(logger, logging.WARNING),
    )
    async def handle_server_command(self, client, command):
        if command == CREATE_CLIENT_CHANNEL:
            remote_reader, remote_writer = await asyncio.open_connection(self.remote_host, self.remote_port)
            readers.add(remote_reader)
            writers.add(remote_writer)
            logger.debug(f"Connected client {client} to remote host {(self.remote_host, self.remote_port)}")
            message = CREATE_CLIENT_CHANNEL + client
            logger.debug(f"handle_server_command: Writing {len(message)} bytes: {readable_hex(message)}")
            remote_writer.write(self.encryptor.encrypt(message, additional_data=self.key_id.encode()))
            await remote_writer.drain()

            # On Windows, we go over TCP because UDS are not supported yet.
            # Check on gRPC every once in a while for eventual support.
            if sys.platform == "win32":
                address_parts = self.local_address.split(":")
                grpc_reader, grpc_writer = await asyncio.open_connection(address_parts[0], address_parts[1])
            # On Linux, we can use Unix Domain Sockets.
            else:
                grpc_reader, grpc_writer = await asyncio.open_unix_connection(urlparse(self.local_address).path)

            readers.add(grpc_reader)
            writers.add(grpc_writer)
            logger.debug(f"Connected client {client} to grpc host {self.local_address}")

            await self.create_sync_task(
                client, grpc_reader, remote_writer, encrypted_writer=True, task_name=f"{client.decode()}: grpc->remote"
            )
            await self.create_sync_task(
                client, remote_reader, grpc_writer, encrypted_reader=True, task_name=f"{client.decode()}: remote->grpc"
            )

    async def create_sync_task(
        self, client, reader, writer, encrypted_writer=False, encrypted_reader=False, task_name=""
    ):
        task = self.sync_data(client, reader, writer, encrypted_writer, encrypted_reader, task_name)
        if sys.version_info >= (3, 8):
            asyncio.create_task(task, name=task_name)
        else:
            asyncio.create_task(task)

    async def sync_data(
        self,
        client,
        reader: StreamReader,
        writer: StreamWriter,
        encrypted_writer=False,
        encrypted_reader=False,
        task_name: str = None,
    ):
        log_prefix = "" if not task_name else task_name + ": "
        try:
            while not reader.at_eof():
                if encrypted_reader:
                    length = await reader.readexactly(4)
                    data = await reader.readexactly(int.from_bytes(length, "big"))
                    if len(data):
                        data = self.encryptor.decrypt(data)
                else:
                    data = await reader.read(2048)
                if len(data):
                    logger.debug(f"{log_prefix}Writing {len(data)} bytes: {readable_hex(data)}")
                    if encrypted_writer:
                        data = self.encryptor.encrypt(data)
                    writer.write(data)
                    await writer.drain()
                else:
                    logger.debug(f"{log_prefix}Connection lost: {client}")
                    break
        except (IncompleteReadError, ConnectionLost, ConnectionResetError):
            logger.debug(f"{log_prefix}Connection lost: {client}")
        except BaseException:  # noqa
            logger.exception(f"{log_prefix}Sync data error")
        finally:
            writer.close()
            readers.remove(reader)
            writers.remove(writer)



import functools
import inspect
from typing import Optional, List, Union, Type, Callable

import pandas as pd

from giskard.core.core import DatasetProcessFunctionMeta
from giskard.core.validation import configured_validate_arguments
from giskard.ml_worker.core.savable import RegistryArtifact
from giskard.ml_worker.testing.registry.decorators_utils import (
    validate_arg_type,
    drop_arg,
    make_all_optional_or_suite_input,
    set_return_type,
)
from giskard.ml_worker.testing.registry.registry import get_object_uuid, tests_registry

TransformationFunctionType = Callable[..., Union[pd.Series, pd.DataFrame]]

default_tags = ["transformation"]


class TransformationFunction(RegistryArtifact[DatasetProcessFunctionMeta]):
    func: TransformationFunctionType = None
    row_level: bool = True
    cell_level: bool = False
    params = {}
    is_initialized = False

    @classmethod
    def _get_name(cls) -> str:
        return "transformations"

    def __init__(self, func: Optional[TransformationFunctionType], row_level=True, cell_level=False):
        self.func = func
        self.row_level = row_level
        self.cell_level = cell_level

        test_uuid = get_object_uuid(func)
        meta = tests_registry.get_test(test_uuid)
        if meta is None and func is not None:
            meta = tests_registry.register(
                DatasetProcessFunctionMeta(func, tags=default_tags, type="TRANSFORMATION", cell_level=self.cell_level)
            )
        super().__init__(meta)

    def __call__(self, *args, **kwargs) -> "TransformationFunction":
        self.is_initialized = True
        self.params = kwargs

        for idx, arg in enumerate(args):
            self.params[next(iter([arg.name for arg in self.meta.args.values() if arg.argOrder == idx]))] = arg

        return self

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms the data using the transformation function.

        Args:
            data (Union[pd.Series, pd.DataFrame]): The data to transform.

        Returns:
            Union[pd.Series, pd.DataFrame]: The transformed data.
        """
        if self.cell_level:
            actual_params = {k: v for k, v in self.params.items() if k != "column_name"}

            def apply(row: pd.Series) -> pd.Series:
                row[self.params["column_name"]] = self.func(row[self.params["column_name"]], **actual_params)
                return row

            return data.apply(apply, axis=1)
        elif self.row_level:
            return data.apply(lambda row: self.func(row, **self.params), axis=1)
        else:
            return self.func(data, **self.params)

    @classmethod
    def _get_meta_class(cls):
        return DatasetProcessFunctionMeta


def transformation_function(
    _fn: Union[TransformationFunctionType, Type[TransformationFunction]] = None,
    row_level=True,
    cell_level=False,
    name=None,
    tags: Optional[List[str]] = None,
):
    """
    Decorator that registers a function as a transformation function and returns a TransformationFunction instance.
    It can be used for transforming datasets in a specific way during testing.

    :param _fn: function to decorate. No need to provide this argument, the decorator will automatically take as input the function to decorate.
    :param name: Optional name to use for the function when registering it.
    :param tags: Optional list of tags to use when registering the function.
    :param row_level: Whether to apply the transformation function row-wise (default) or on the full dataframe. If `row_level`
                      is True, the slicing function will receive a row (either a Series or DataFrame),
                      and if False, it will receive the entire dataframe.
    :param cell_level: Whether to apply the transformation function on the cell level. If True, the slicing function
                       will be applied to individual cells instead of rows or the entire dataframe.
    :return: The wrapped function or a new instance of TransformationFunction.
    """

    def inner(func: Union[TransformationFunctionType, Type[TransformationFunction]]) -> TransformationFunction:
        from giskard.ml_worker.testing.registry.registry import tests_registry

        tests_registry.register(
            DatasetProcessFunctionMeta(
                func,
                name=name,
                tags=default_tags if not tags else (default_tags + tags),
                type="TRANSFORMATION",
                cell_level=cell_level,
            )
        )

        if inspect.isclass(func) and issubclass(func, TransformationFunction):
            return func
        return _wrap_transformation_function(func, row_level, cell_level)()

    if callable(_fn):
        return functools.wraps(_fn)(inner(_fn))
    else:
        return inner


def _wrap_transformation_function(original: Callable, row_level: bool, cell_level: bool):
    transformation_fn = functools.wraps(original)(TransformationFunction(original, row_level, cell_level))

    if not cell_level:
        validate_arg_type(transformation_fn, 0, pd.Series if row_level else pd.DataFrame)

    drop_arg(transformation_fn, 0)

    make_all_optional_or_suite_input(transformation_fn)
    set_return_type(transformation_fn, TransformationFunction)

    return configured_validate_arguments(transformation_fn)



from . import titanic_classification, linear_regression


def titanic_df():
    return titanic_classification.get_df()


def titanic():
    return titanic_classification.get_model_and_df()


def titanic_pipeline():
    return titanic_classification.get_pipeline()


def linear_df():
    return linear_regression.get_df()


def linear():
    return linear_regression.get_model_and_df()


def linear_pipeline():
    return linear_regression.get_pipeline()


__all__ = [
    "titanic_df",
    "titanic",
    "titanic_pipeline",
    "linear_df",
    "linear",
    "linear_pipeline",
]



"""Statistical tests"""
import numbers
import numpy as np
from typing import Optional

from giskard import test
from giskard.datasets.base import Dataset
from giskard.ml_worker.testing.test_result import TestResult, TestMessage, TestMessageLevel
from giskard.ml_worker.testing.utils import validate_classification_label
from giskard.ml_worker.testing.registry.slicing_function import SlicingFunction
from giskard.models.base import BaseModel
from giskard.ml_worker.testing.utils import check_slice_not_empty


@test(name="Right Label", tags=["heuristic", "classification"])
@validate_classification_label
def test_right_label(
    model: BaseModel,
    dataset: Dataset,
    classification_label: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.5,
) -> TestResult:
    """
    Summary: Test if the model returns the right classification label for a slice

    Description: The test is passed when the percentage of rows returning the right
    classification label is higher than the threshold in a given slice

    Example: For a credit scoring model, the test is passed when more than 50%
    of people with high-salaries are classified as “non default”


    Args:
      model(BaseModel):
          Model used to compute the test
      dataset(Dataset):
          Dataset used to compute the test
      classification_label(str):
          Classification label you want to test
      slicing_function(Optional[SlicingFunction]):
          Slicing function to be applied on the dataset
      threshold(float):
          Threshold for the percentage of passed rows

    Returns:
      actual_slices_size:
          Length of dataset tested
      metrics:
          The ratio of rows with the right classification label over the total of rows in the slice
      passed:
          TRUE if passed_ratio > threshold
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_right_label")

    prediction_results = model.predict(dataset).prediction

    passed_idx = dataset.df.loc[prediction_results == classification_label].index.values

    passed_ratio = len(passed_idx) / len(dataset)
    return TestResult(
        actual_slices_size=[len(dataset)],
        metric=passed_ratio,
        passed=bool(passed_ratio > threshold),
    )


@test(name="Output in range", tags=["heuristic", "classification", "regression"])
@validate_classification_label
def test_output_in_range(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    classification_label: Optional[str] = None,
    min_range: float = 0.3,
    max_range: float = 0.7,
    threshold: float = 0.5,
) -> TestResult:
    """
    Summary: Test if the model output belongs to the right range for a slice

    Description: - The test is passed when the ratio of rows in the right range inside the
    slice is higher than the threshold.

    For classification: Test if the predicted probability for a given classification label
    belongs to the right range for a dataset slice

    For regression : Test if the predicted output belongs to the right range for a dataset slice

    Example :
    For Classification: For a credit scoring model, the test is passed when more than 50% of
    people with high wage have a probability of defaulting between 0 and 0.1

    For Regression : The predicted Sale Price of a house in the city falls in a particular range


    Args:
        model(BaseModel):
            Model used to compute the test
        dataset(Dataset):
            Dataset used to compute the test
        slicing_function(Optional[SlicingFunction]):
            Slicing function to be applied on the dataset
        classification_label(Optional[str]):
            Optional. Classification label you want to test
        min_range(float):
            Minimum probability of occurrence of classification label
        max_range(float):
            Maximum probability of occurrence of classification label
        threshold(float):
            Threshold for the percentage of passed rows

    Returns:
        actual_slices_size:
            Length of dataset tested
        metrics:
            The proportion of rows in the right range inside the slice
        passed:
            TRUE if metric > threshold
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_output_in_range")

    prediction_results = model.predict(dataset)

    if model.is_regression:
        output = prediction_results.raw_prediction

    elif model.is_classification:
        output = prediction_results.all_predictions[classification_label]

    else:
        raise ValueError(f"Prediction task is not supported: {model.meta.model_type}")

    passed_idx = dataset.df.loc[(output <= max_range) & (output >= min_range)].index.values

    passed_ratio = len(passed_idx) / len(dataset)

    return TestResult(
        actual_slices_size=[len(dataset)],
        metric=passed_ratio,
        passed=bool(passed_ratio >= threshold),
    )


@test(name="Disparate impact", tags=["heuristic", "classification"])
def test_disparate_impact(
    model: BaseModel,
    dataset: Dataset,
    protected_slicing_function: SlicingFunction,
    unprotected_slicing_function: SlicingFunction,
    positive_outcome: str,
    slicing_function: Optional[SlicingFunction] = None,
    min_threshold: float = 0.8,
    max_threshold: float = 1.25,
) -> TestResult:
    """
    Summary: Tests if the model is biased more towards an unprotected slice of the dataset over a protected slice.
    Note that this test reflects only a possible bias in the model while being agnostic to any biaas in the dataset
    it trained on. The Disparate Impact (DI) is only valid for classification models and is computed as the ratio
    between the average count of correct predictions for the protected slice over the unprotected one given a
    certain positive_outcome.

    Description: Calculate the Disparate Impact between a protected and unprotected slice of a dataset. Otherwise
    known as the "80 percent" rule, the Disparate Impact determines if a model was having an "adverse impact" on a
    protected (or minority in some cases) group.

    Example: The rule was originally based on the rates at which job applicants were hired. For example, if XYZ
    Company hired 50 percent of the men applying for work in a predominantly male occupation while hiring only 20
    percent of the female applicants, one could look at the ratio of those two hiring rates to judge whether there
    might be a discrimination problem. The ratio of 20:50 means that the rate of hiring for female applicants is
    only 40 percent of the rate of hiring for male applicants. That is, 20 divided by 50 equals
    0.40, which is equivalent to 40 percent. Clearly, 40 percent is well below the 80 percent that was arbitrarily
    set as an acceptable difference in hiring rates. Therefore, in this example, XYZ Company could have been called
    upon to prove that there was a legitimate reason for hiring men at a rate so much higher than the rate of hiring
    women.

    Args:
          model(BaseModel):
              Model used to compute the test
          dataset(Dataset):
              Dataset used to compute the test
          protected_slicing_function(SlicingFunction):
              Slicing function that defines the protected group from the full dataset given
          unprotected_slicing_function(SlicingFunction):
              Slicing function that defines the unprotected group from the full dataset given
          positive_outcome(str):
              The target value that is considered a positive outcome in the dataset
          slicing_function(Optional[SlicingFunction]):
              Slicing function to be applied on the dataset
          min_threshold(float):
              Threshold below which the DI test is considered to fail, by default 0.8
          max_threshold(float):
              Threshold above which the DI test is considered to fail, by default 1.25

    Returns:
          metric:
              The disparate impact ratio
          passed:
              TRUE if the disparate impact ratio > min_threshold && disparate impact ratio < max_threshold
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_disparate_impact")

    # Try to automatically cast `positive_outcome` to the right type
    if isinstance(model.meta.classification_labels[0], numbers.Number):
        try:
            positive_outcome = float(positive_outcome)
        except ValueError:
            pass

    if positive_outcome not in list(model.meta.classification_labels):
        raise ValueError(
            f"The positive outcome chosen {positive_outcome} is not part of the dataset target values {list(model.meta.classification_labels)}."
        )

    protected_ds = dataset.slice(protected_slicing_function)
    unprotected_ds = dataset.slice(unprotected_slicing_function)

    if protected_ds.df.equals(unprotected_ds.df):
        raise ValueError(
            "The protected and unprotected datasets are equal. Please check that you chose different slices."
        )

    positive_idx = list(model.meta.classification_labels).index(positive_outcome)

    protected_predictions = np.squeeze(model.predict(protected_ds).raw_prediction == positive_idx)
    unprotected_predictions = np.squeeze(model.predict(unprotected_ds).raw_prediction == positive_idx)

    protected_proba = np.count_nonzero(protected_predictions) / len(protected_ds.df)
    unprotected_proba = np.count_nonzero(unprotected_predictions) / len(unprotected_ds.df)
    disparate_impact_score = protected_proba / unprotected_proba

    messages = [
        TestMessage(
            type=TestMessageLevel.INFO, text=f"min_threshold = {min_threshold}, max_threshold = {max_threshold}"
        )
    ]

    return TestResult(
        metric=disparate_impact_score,
        passed=bool((disparate_impact_score > min_threshold) * (disparate_impact_score < max_threshold)),
        messages=messages,
    )



import logging
import os
from dataclasses import dataclass
from pathlib import Path

from git import Repo

from giskard.settings import expand_env_var, settings

logger = logging.getLogger(__name__)
udf_root = Path(expand_env_var(settings.home)) / "cache" / "udf"


@dataclass
class GitConfig:
    repo: str
    user: str
    pwd: str


def load_config() -> GitConfig:
    return GitConfig(
        repo=os.getenv("GIT_REPOSITORY"),
        user=os.getenv("GIT_USER"),
        pwd=os.getenv("GIT_PASSWORD"),
    )


def init_udf_repository(udf_enabled: bool = False) -> bool:
    # Experimental feature, disabled by default
    if not udf_enabled:
        return False

    try:
        config = load_config()

        if config.repo is None:
            logger.warning(
                "UDF repository is not available because the 'GIT_REPOSITORY' environment variable is not set"
            )
            return False

        if udf_root.exists():
            update_udf_repository()
        else:
            clone_udf_repository(config)
    except Exception as e:
        logger.exception("Failed to initialize UDF repo", e)
        return False
    return True


def update_udf_repository():
    repo = Repo(udf_root)

    logger.info("Repository already existing, fetching updates")
    origin = repo.remotes.origin
    origin.pull()

    repo.close()


def clone_udf_repository(config: GitConfig):
    logger.info(f"cloning repository: {config.repo}")

    host = config.repo

    protocol = "http"
    if not host.startswith("http://"):
        protocol = "https"

    host = f"{protocol}://{config.user}:{config.pwd}@{config.repo.replace('http://', '').replace('https://', '')}"

    repo = Repo.clone_from(host, udf_root)
    repo.close()


udf_repo_available = init_udf_repository()



import collections
import importlib
from pathlib import Path
from typing import Union, Literal, get_args, Optional

import mlflow
import numpy as np
import pandas as pd
import torch
import yaml
from torch.utils.data import DataLoader
from torch.utils.data import Dataset as torch_dataset

from giskard.core.core import ModelType
from giskard.models.base import MLFlowBasedModel
from ..utils import map_to_tuples

TorchDType = Literal[
    "float32",
    "float",
    "float64",
    "double",
    "complex64",
    "cfloat",
    "float16",
    "half",
    "bfloat16",
    "uint8",
    "int8",
    "int16",
    "short",
    "int32",
    "int",
    "int64",
    "long",
    "bool",
]


def string_to_torch_dtype(torch_dtype_string: TorchDType):
    try:
        return getattr(importlib.import_module("torch"), torch_dtype_string)
    except AttributeError:
        raise ValueError(
            f"Incorrect torch dtype specified: {torch_dtype_string}, " f"available values are: {get_args(TorchDType)}"
        )


class TorchMinimalDataset(torch_dataset):
    def __init__(self, df: pd.DataFrame, torch_dtype: TorchDType = "float32"):
        self.entries = df
        self.torch_dtype = torch_dtype

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, idx):
        return torch.tensor(self.entries.iloc[idx].to_numpy(), dtype=string_to_torch_dtype(self.torch_dtype))


class PyTorchModel(MLFlowBasedModel):
    """
    A wrapper class for PyTorch models that extends the functionality of the
    MLFlowBasedModel class.

    Attributes:
        iterate_dataset (bool, optional): Whether to iterate over the dataset for prediction. Defaults to True.
        device (str): The device to run the model on.
        torch_dtype (TorchDType): The data type to be used for input tensors.
    """

    def __init__(
        self,
        model,
        model_type: ModelType,
        torch_dtype: TorchDType = "float32",
        device="cpu",
        name: Optional[str] = None,
        data_preprocessing_function=None,
        model_postprocessing_function=None,
        feature_names=None,
        classification_threshold=0.5,
        classification_labels=None,
        iterate_dataset=True,
        **kwargs,
    ) -> None:
        super().__init__(
            model=model,
            model_type=model_type,
            name=name,
            data_preprocessing_function=data_preprocessing_function,
            model_postprocessing_function=model_postprocessing_function,
            feature_names=feature_names,
            classification_threshold=classification_threshold,
            classification_labels=classification_labels,
            **kwargs,
        )

        self.device = device
        self.torch_dtype = torch_dtype
        self.iterate_dataset = iterate_dataset

    @classmethod
    def load_model(cls, local_dir):
        return mlflow.pytorch.load_model(local_dir)

    def save_model(self, local_path, mlflow_meta: mlflow.models.Model):
        mlflow.pytorch.save_model(self.model, path=local_path, mlflow_model=mlflow_meta)

    def _get_predictions_from_iterable(self, data):
        # Fault tolerance: try to convert to the right format in special cases
        if isinstance(data, pd.DataFrame):
            data = TorchMinimalDataset(data, self.torch_dtype)
        elif isinstance(data, DataLoader):
            data = _get_dataset_from_dataloader(data)

        # Create the data iterator
        try:
            data_iter = iter(data)
        except TypeError as err:
            raise ValueError(
                f"The data exposed to your model must be iterable (instead, we got type={type(data)}). "
                "Make sure that your data or your `data_preprocessing_function` outputs one of the following:\n"
                "- pandas.DataFrame\n- torch.utils.data.Dataset\n- iterable with elements that are compatible with your model"
            ) from err

        try:
            with torch.no_grad():
                return torch.cat([self.model(*entry) for entry in map_to_tuples(data_iter)])
        except ValueError as err:
            raise ValueError(
                "Running your model prediction on one element of your dataset returned an error.\n"
                "Please check that your `data_preprocessing_function` returns an iterable of objects "
                "that are valid inputs for your model."
            ) from err

    def _get_predictions_from_object(self, data):
        try:
            return self.model(data)
        except ValueError as err:
            raise ValueError(
                "Running your model prediction returned an error.\n"
                "Since you specified `iter_dataset=False`, please check that your `data_preprocessing_function` "
                "returns an object that can be passed as input for your model. "
            ) from err

    def model_predict(self, data):
        self.model.to(self.device)
        self.model.eval()

        if self.iterate_dataset:
            predictions = self._get_predictions_from_iterable(data)
        else:
            predictions = self._get_predictions_from_object(data)

        return predictions

    def _convert_to_numpy(self, raw_predictions):
        if isinstance(raw_predictions, torch.Tensor):
            return raw_predictions.detach().numpy()

        return np.asarray(raw_predictions)

    def save_pytorch_meta(self, local_path):
        with open(Path(local_path) / "giskard-model-pytorch-meta.yaml", "w") as f:
            yaml.dump(
                {
                    "device": self.device,
                    "torch_dtype": self.torch_dtype,
                    "iterate_dataset": self.iterate_dataset,
                },
                f,
                default_flow_style=False,
            )

    def save(self, local_path: Union[str, Path]) -> None:
        super().save(local_path)
        self.save_pytorch_meta(local_path)

    @classmethod
    def load(cls, local_dir, **kwargs):
        pytorch_meta_file = Path(local_dir) / "giskard-model-pytorch-meta.yaml"
        if pytorch_meta_file.exists():
            with open(pytorch_meta_file) as f:
                pytorch_meta = yaml.load(f, Loader=yaml.Loader)
                kwargs["device"] = pytorch_meta["device"]
                kwargs["torch_dtype"] = pytorch_meta["torch_dtype"]
                kwargs["iterate_dataset"] = pytorch_meta.get("iterate_dataset")
                return super().load(local_dir, **kwargs)
        else:
            raise ValueError(
                f"Cannot load model ({cls.__module__}.{cls.__name__}), " f"{pytorch_meta_file} file not found"
            )


def _get_dataset_from_dataloader(dl: DataLoader):
    if not isinstance(dl.dataset, collections.defaultdict):
        return dl.dataset

    raise ValueError(
        f"We tried to infer the torch.utils.data.Dataset from your DataLoader. \n \
                    The type we found was {dl.dataset} which we don't support. Please provide us \n \
                    with a different iterable as output of your data_preprocessing_function."
    )


def _convert_to_numpy(self, predictions):
    return torch.cat(predictions).detach().squeeze(0).numpy()



import pytest

from giskard import test, scan
from giskard.core.suite import Suite, SuiteInput
from giskard.datasets.base import Dataset
from giskard.ml_worker.testing.registry.slicing_function import SlicingFunction, slicing_function
from giskard.ml_worker.testing.test_result import TestMessageLevel
from giskard.models.base import BaseModel
from giskard.testing.tests.performance import test_auc, test_f1, test_diff_f1


@test()
def _test_a_greater_b(a: int, b: int):
    return a > b


def test_a_greater_b_fail():
    passed, _ = Suite().add_test(_test_a_greater_b(1, 2)).run()
    assert not passed


def test_a_greater_b_pass():
    passed, _ = Suite().add_test(_test_a_greater_b(2, 1)).run()
    assert passed


def test_missing_arg():
    with pytest.raises(Exception, match="Missing 1 required parameters: {'b': <class 'int'>}"):
        Suite().add_test(_test_a_greater_b(a=2)).run()


def test_missing_args():
    with pytest.raises(Exception, match="Missing 2 required parameters: {'a': <class 'int'>, 'b': <class 'int'>}"):
        Suite().add_test(_test_a_greater_b()).run()


def test_missing_arg_one_global():
    with pytest.raises(Exception, match="Missing 1 required parameters: {'b': <class 'int'>}"):
        Suite().add_test(_test_a_greater_b()).run(a=2)


def test_all_global():
    passed, _ = Suite().add_test(_test_a_greater_b()).run(a=2, b=1)
    assert passed


def test_multiple(german_credit_data: Dataset, german_credit_model: BaseModel):
    assert (
        Suite()
        .add_test(test_auc(threshold=0.2))
        .add_test(test_f1(threshold=0.2))
        .run(dataset=german_credit_data, model=german_credit_model)[0]
    )


def test_all_inputs_exposed_and_shared(german_credit_data, german_credit_model):
    assert (
        Suite()
        .add_test(test_auc())
        .add_test(test_f1())
        .add_test(_test_a_greater_b())
        .run(dataset=german_credit_data, model=german_credit_model, threshold=0.2, a=2, b=1)[0]
    )


def test_shared_input(german_credit_model: BaseModel, german_credit_data: Dataset):
    last_half = german_credit_data.slice(SlicingFunction(lambda df: df.tail(len(df) // 2), row_level=False))

    shared_input = SuiteInput("dataset", Dataset)

    assert (
        Suite()
        .add_test(test_auc(dataset=shared_input, threshold=0.2))
        .add_test(test_f1(dataset=shared_input, threshold=0.2))
        .add_test(test_diff_f1(threshold=0.2, actual_dataset=shared_input))
        .run(model=german_credit_model, dataset=german_credit_data, reference_dataset=last_half)[0]
    )


def test_multiple_execution_of_same_test(german_credit_data: Dataset, german_credit_model: BaseModel):
    first_half = german_credit_data.slice(SlicingFunction(lambda df: df.head(len(df) // 2), row_level=False))
    last_half = german_credit_data.slice(SlicingFunction(lambda df: df.tail(len(df) // 2), row_level=False))

    shared_input = SuiteInput("dataset", Dataset)

    result = (
        Suite()
        .add_test(test_auc(dataset=shared_input, threshold=0.2))
        .add_test(test_auc(dataset=shared_input, threshold=0.25))
        .add_test(test_auc(dataset=shared_input, threshold=0.3))
        .run(
            model=german_credit_model,
            dataset=german_credit_data,
            actual_dataset=first_half,
            reference_dataset=last_half,
        )
    )

    assert result[0]
    assert len(result[1]) == 3


def test_giskard_test_class(german_credit_data: Dataset, german_credit_model: BaseModel):
    shared_input = SuiteInput("dataset", Dataset)

    assert (
        Suite()
        .add_test(test_auc(dataset=shared_input, threshold=0.2))
        .run(model=german_credit_model, dataset=german_credit_data)[0]
    )


def test_execution_error(german_credit_data: Dataset, german_credit_model: BaseModel):
    shared_input = SuiteInput("dataset", Dataset)

    @slicing_function(row_level=False)
    def empty_slice(x):
        return x.iloc[[]]

    result = Suite().add_test(test_f1(dataset=shared_input, threshold=0.2, slicing_function=empty_slice)) \
        .add_test(test_f1(dataset=shared_input, threshold=0.2)) \
        .run(model=german_credit_model, dataset=german_credit_data)

    assert result[0] is False
    assert result[1][0][1].passed is False
    assert result[1][0][1].is_error is True
    assert result[1][0][1].messages[0].type is TestMessageLevel.ERROR
    assert 'The sliced dataset in test_f1 is empty.' in result[1][0][1].messages[0].text
    assert result[1][1][1].passed is True
    assert result[1][1][1].is_error is False


def test_save_suite(german_credit_data: Dataset, german_credit_model: BaseModel):
    # with MockedClient() as (client, mr):
    #     Suite().add_test(test_auc(threshold=0.2, dataset=german_credit_data)).add_test(
    #         test_f1(threshold=0.2, dataset=german_credit_data)
    #     ).upload(client, "test_project_key")
    scan(german_credit_model, german_credit_data)


# def test_save_suite_real(german_credit_data: Dataset, german_credit_model: BaseModel):
#     from giskard.client.giskard_client import GiskardClient
#     client = GiskardClient("http://localhost:9000",
#                            "")
#
#     german_credit_data.upload(client, 'test')
#     german_credit_model.upload(client, 'test')
#
#     Suite(name="Test Suite 1") \
#         .add_test(test_auc, threshold=0.2, dataset=german_credit_data) \
#         .add_test(test_f1, threshold=0.2, dataset=german_credit_data) \
#         .save(client, 'credit')



from typing import Sequence
import pandas as pd

from ...testing.tests.calibration import _default_overconfidence_threshold

from ...testing.tests.calibration import _calculate_overconfidence_score

from ...ml_worker.testing.registry.slicing_function import SlicingFunction
from ...models.base import BaseModel
from ...datasets import Dataset
from ..decorators import detector
from ..common.loss_based_detector import LossBasedDetector
from .issues import CalibrationIssue, CalibrationIssueInfo, OverconfidenceIssue
from ..logger import logger


@detector(name="overconfidence", tags=["overconfidence", "classification"])
class OverconfidenceDetector(LossBasedDetector):
    def __init__(self, threshold=0.10, p_threshold=None, method="tree"):
        self.threshold = threshold
        self.p_threshold = p_threshold
        self.method = method

    @property
    def _numerical_slicer_method(self):
        return self.method

    def run(self, model: BaseModel, dataset: Dataset):
        if not model.is_classification:
            raise ValueError("Overconfidence bias detector only works for classification models.")

        return super().run(model, dataset)

    def _calculate_loss(self, model: BaseModel, dataset: Dataset) -> pd.DataFrame:
        loss = _calculate_overconfidence_score(model, dataset).to_frame(self.LOSS_COLUMN_NAME)
        return loss[loss[self.LOSS_COLUMN_NAME] > 0]

    def _find_issues(
        self,
        slices: Sequence[SlicingFunction],
        model: BaseModel,
        dataset: Dataset,
        meta: pd.DataFrame,
    ) -> Sequence[CalibrationIssue]:
        # Add the loss column to the dataset
        dataset_with_meta = Dataset(
            dataset.df.join(meta, how="left"),
            target=dataset.target,
            column_types=dataset.column_types,
        )
        # For performance
        dataset_with_meta.load_metadata_from_instance(dataset.column_meta)

        p_threshold = self.p_threshold or _default_overconfidence_threshold(model)
        logger.info(f"{self.__class__.__name__}: Using overconfidence threshold = {p_threshold}")

        reference_rate = (dataset_with_meta.df[self.LOSS_COLUMN_NAME].dropna() > p_threshold).mean()

        issues = []
        for slice_fn in slices:
            sliced_dataset = dataset_with_meta.slice(slice_fn)

            slice_rate = (sliced_dataset.df[self.LOSS_COLUMN_NAME].dropna() > p_threshold).mean()
            fail_idx = sliced_dataset.df[(sliced_dataset.df[self.LOSS_COLUMN_NAME] > p_threshold)].index
            relative_delta = (slice_rate - reference_rate) / reference_rate

            if relative_delta > self.threshold:
                level = "major" if relative_delta > 2 * self.threshold else "medium"
                issues.append(
                    OverconfidenceIssue(
                        model,
                        dataset,
                        level,
                        CalibrationIssueInfo(
                            slice_fn=slice_fn,
                            slice_size=len(sliced_dataset),
                            metric_value_slice=slice_rate,
                            metric_value_reference=reference_rate,
                            loss_values=meta[self.LOSS_COLUMN_NAME],
                            fail_idx=fail_idx,
                            threshold=self.threshold,
                            p_threshold=p_threshold,
                        ),
                    )
                )

        return issues



import numpy as np
import scipy.stats as stats


def paired_t_test(population_1, population_2, alternative, critical_quantile):
    """
    Computes the result of the paired t-test given 2 groups of observations and a level of significance

    p_value > quantile ==> we support the null hypothesis
    p_value < quantile ==> we support the alternative hypothesis

    less ==> alternative: mean(pop1) < mean(pop2)
    greater ==> alternative: mean(pop1) > mean(pop2)
    """

    if alternative not in ["less", "greater"]:
        raise ValueError("Incorrect alternative hypothesis! It has to be one of the following: ['less', 'greater']")

    if np.array_equal(population_1, population_2):
        p_value = 1.0
    else:
        _, p_value = stats.ttest_rel(population_1, population_2, alternative=alternative)

    return p_value <= critical_quantile, p_value


def equivalence_t_test(population_1, population_2, window_size, critical_quantile):
    """
    Computes the equivalence test to show that mean 2 - window_size/2 < mean 1 < mean 2 + window_size/2

    Inputs:
        - population_1, np.array 1D: an array of 1 dimension with the observations of 1st group
        - population_2, np.array 1D: an array of 1 dimension with the observations of 2nd group
        - window_size, float: small value that determines the maximal difference that can be between means
        - critical_quantile, float: the level of significance, is the 1-q quantile of the distribution

    Output, bool: True if the null is rejected and False if there is not enough evidence
    """
    population_2_up = population_2 + window_size / 2.0
    population_2_low = population_2 - window_size / 2.0

    if np.array_equal(population_1, population_2):
        p_value_low, p_value_up = 0.0, 0.0
    else:
        p_value_up = stats.ttest_rel(population_1, population_2_up, alternative="less")[1]
        p_value_low = stats.ttest_rel(population_1, population_2_low, alternative="greater")[1]

    test_up = p_value_up < critical_quantile
    test_low = p_value_low < critical_quantile

    if test_low and test_up:
        print("null hypothesis rejected at a level of significance", critical_quantile)
        return True, max(p_value_low, p_value_up)
    else:
        print("not enough evidence to reject null hypothesis")
        return False, max(p_value_low, p_value_up)


def paired_wilcoxon(population_1, population_2, alternative, critical_quantile):
    """
    Computes the result of the Wilcoxon rank-sum given 2 groups of observations and a level of significance

    p_value > quantile ==> we support the null hypothesis
    p_value < quantile ==> we support the alternative hypothesis

    less ==> alternative: mean(pop1) < mean(pop2)
    greater ==> alternative: mean(pop1) > mean(pop2)
    """

    if alternative not in ["less", "greater"]:
        raise ValueError("Incorrect alternative hypothesis! It has to be one of the following: ['less', 'greater']")

    if np.array_equal(population_1, population_2):
        p_value = 1.0
    else:
        _, p_value = stats.wilcoxon(population_1, population_2, alternative=alternative)

    return p_value <= critical_quantile, p_value


def equivalence_wilcoxon(population_1, population_2, window_size, critical_quantile):
    """
    Computes the equivalence test to show that mean 2 - window_size/2 < mean 1 < mean 2 + window_size/2

    Inputs:
        - population_1, np.array 1D: an array of 1 dimension with the observations of 1st group
        - population_2, np.array 1D: an array of 1 dimension with the observations of 2nd group
        - threshold, float: small value that determines the maximal difference that can be between means
        - quantile, float: the level of significance, is the 1-q quantile of the distribution

    Output, bool: True if the null is rejected and False if there is not enough evidence
    """
    population_2_up = population_2 + window_size / 2.0
    population_2_low = population_2 - window_size / 2.0

    if np.array_equal(population_1, population_2):
        p_value_low, p_value_up = 0.0, 0.0
    else:
        p_value_up = stats.wilcoxon(population_1, population_2_up, alternative="less")[1]
        p_value_low = stats.wilcoxon(population_1, population_2_low, alternative="greater")[1]

    test_up = p_value_up < critical_quantile
    test_low = p_value_low < critical_quantile

    if test_low and test_up:
        print("null hypothesis rejected at a level of significance", critical_quantile)
        return True, max(p_value_low, p_value_up)
    else:
        print("not enough evidence to reject null hypothesis")
        return False, max(p_value_low, p_value_up)



from abc import abstractmethod
import numpy as np
import pandas as pd
from typing import Sequence, Optional

from ..llm.utils import LLMImportError

from .text_transformations import TextTransformation
from ..issues import Issue
from ...datasets.base import Dataset
from ...models.base import BaseModel
from ..registry import Detector
from .issues import RobustnessIssue, RobustnessIssueInfo
from ..logger import logger


class BaseTextPerturbationDetector(Detector):
    _issue_cls = RobustnessIssue

    def __init__(
        self,
        transformations: Optional[Sequence[TextTransformation]] = None,
        threshold: Optional[float] = None,
        output_sensitivity=None,
        num_samples: Optional[int] = None,
    ):
        self.transformations = transformations
        self.threshold = threshold
        self.num_samples = num_samples
        self.output_sensitivity = output_sensitivity

    def run(self, model: BaseModel, dataset: Dataset) -> Sequence[Issue]:
        transformations = self.transformations or self._get_default_transformations(model, dataset)
        features = [
            col
            for col, col_type in dataset.column_types.items()
            if col_type == "text" and pd.api.types.is_string_dtype(dataset.df[col].dtype)
        ]

        # Only analyze the model features
        if model.meta.feature_names:
            features = [f for f in features if f in model.meta.feature_names]

        logger.info(
            f"{self.__class__.__name__}: Running with transformations={[t.name for t in transformations]} "
            f"threshold={self.threshold} output_sensitivity={self.output_sensitivity} num_samples={self.num_samples}"
        )

        issues = []
        for transformation in transformations:
            issues.extend(self._detect_issues(model, dataset, transformation, features))

        return [i for i in issues if i is not None]

    @abstractmethod
    def _get_default_transformations(self, model: BaseModel, dataset: Dataset) -> Sequence[TextTransformation]:
        ...

    def _detect_issues(
        self,
        model: BaseModel,
        dataset: Dataset,
        transformation: TextTransformation,
        features: Sequence[str],
    ) -> Sequence[Issue]:
        num_samples = self.num_samples or _get_default_num_samples(model)
        output_sensitivity = self.output_sensitivity or _get_default_output_sensitivity(model)
        threshold = self.threshold or _get_default_threshold(model)

        issues = []
        # @TODO: integrate this with Giskard metamorphic tests already present
        for feature in features:
            transformation_fn = transformation(column=feature)
            transformed = dataset.transform(transformation_fn)

            # Select only the records which were changed
            changed_idx = dataset.df.index[transformed.df[feature] != dataset.df[feature]]

            if changed_idx.empty:
                continue

            # Select a random subset of the changed records
            if len(changed_idx) > num_samples:
                rng = np.random.default_rng(747)
                changed_idx = changed_idx[rng.choice(len(changed_idx), num_samples, replace=False)]

            original_data = Dataset(
                dataset.df.loc[changed_idx],
                target=dataset.target,
                column_types=dataset.column_types,
                validation=False,
            )
            perturbed_data = Dataset(
                transformed.df.loc[changed_idx],
                target=dataset.target,
                column_types=dataset.column_types,
                validation=False,
            )

            # Calculate predictions
            original_pred = model.predict(original_data)
            perturbed_pred = model.predict(perturbed_data)

            if model.is_classification:
                passed = original_pred.raw_prediction == perturbed_pred.raw_prediction
            elif model.is_regression:
                rel_delta = _relative_delta(perturbed_pred.raw_prediction, original_pred.raw_prediction)
                passed = np.abs(rel_delta) < output_sensitivity
            elif model.is_text_generation:
                try:
                    import evaluate
                except ImportError as err:
                    raise LLMImportError() from err

                scorer = evaluate.load("bertscore")
                score = scorer.compute(
                    predictions=perturbed_pred.prediction,
                    references=original_pred.prediction,
                    model_type="distilbert-base-multilingual-cased",
                    idf=True,
                )
                passed = np.array(score["f1"]) > 1 - output_sensitivity
            else:
                raise NotImplementedError("Only classification, regression, or text generation models are supported.")

            pass_ratio = passed.mean()
            fail_ratio = 1 - pass_ratio
            logger.info(
                f"{self.__class__.__name__}: Testing `{feature}` for perturbation `{transformation.name}`\tFail rate: {fail_ratio:.3f}"
            )

            if fail_ratio >= threshold:
                info = RobustnessIssueInfo(
                    feature=feature,
                    fail_ratio=fail_ratio,
                    transformation_fn=transformation_fn,
                    perturbed_data_slice=perturbed_data,
                    perturbed_data_slice_predictions=perturbed_pred,
                    fail_data_idx=original_data.df[~passed].index.values,
                    threshold=threshold,
                    output_sensitivity=output_sensitivity,
                )
                issue = self._issue_cls(
                    model,
                    dataset,
                    level="major" if fail_ratio >= 2 * threshold else "medium",
                    info=info,
                )
                issues.append(issue)

        return issues


def _relative_delta(actual, reference):
    return (actual - reference) / reference


def _get_default_num_samples(model) -> int:
    if model.is_text_generation:
        return 10

    return 1_000


def _get_default_output_sensitivity(model) -> float:
    if model.is_text_generation:
        return 0.15

    return 0.05


def _get_default_threshold(model) -> float:
    if model.is_text_generation:
        return 0.10

    return 0.05



from typing import Any, Callable, Optional, Iterable

import numpy
import pandas as pd

from giskard.core.core import ModelType
from giskard.core.validation import configured_validate_arguments
from giskard.models.base import CloudpickleBasedModel


class PredictionFunctionModel(CloudpickleBasedModel):
    @configured_validate_arguments
    def __init__(
        self,
        model: Callable,
        model_type: ModelType,
        data_preprocessing_function: Callable[[pd.DataFrame], Any] = None,
        model_postprocessing_function: Callable[[Any], Any] = None,
        name: Optional[str] = None,
        feature_names: Optional[Iterable] = None,
        classification_threshold: Optional[float] = 0.5,
        classification_labels: Optional[Iterable] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            model,
            model_type,
            data_preprocessing_function,
            model_postprocessing_function,
            name,
            feature_names,
            classification_threshold,
            classification_labels,
            **kwargs,
        )

    def model_predict(self, df: pd.DataFrame) -> numpy.ndarray:
        return self.model(df)



import os
import re

import pkg_resources
from setuptools import setup, Command
from setuptools.command.build_py import build_py


class GrpcTool(Command):
    user_options = []
    out_path = "giskard/ml_worker/generated"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def fix_paths(path):
        for dirpath, dirs, files in os.walk(path):
            for filename in files:
                print(f"Generated file: {dirpath}/{filename}")
                if not filename.endswith(".py"):
                    continue
                fname = os.path.join(dirpath, filename)
                with open(fname) as rfile:
                    content = rfile.read()
                    content = re.sub(
                        "^import ([^\\. ]*?_pb2)",
                        "import giskard.ml_worker.generated.\g<1>",
                        content,
                        flags=re.MULTILINE,
                    )
                    with open(fname, "w") as wfile:
                        wfile.write(content)
                    print(f"Fixed {fname}")

    def run(self):
        print("Running grpc_tools.protoc")
        import grpc_tools.protoc

        os.makedirs(self.out_path, exist_ok=True)
        proto_path = "../common/proto"
        proto_include = pkg_resources.resource_filename("grpc_tools", "_proto")

        cmd = [
            "grpc_tools.protoc",
            f"-I{proto_include}",
            f"-I{proto_path}",
            f"--python_out={self.out_path}",
            f"--grpc_python_out={self.out_path}",
        ]
        try:
            import importlib

            importlib.import_module("mypy_protobuf")
            cmd += [f"--mypy_out={self.out_path}", f"--mypy_grpc_out={self.out_path}"]
        except Exception:
            print("mypy-protobuf isn't installed or 'import mypy_protobuf' didn't work")
        cmd.append("ml-worker.proto")
        print(f"Running: {' '.join(cmd)}")
        grpc_tools.protoc.main(cmd)

        self.fix_paths(self.out_path)


class BuildPyCommand(build_py):
    def finalize_options(self) -> None:
        super().finalize_options()
        self.run_command("grpc")


setup(
    cmdclass={"build_py": BuildPyCommand, "grpc": GrpcTool},
)



from collections import defaultdict
from typing import Sequence, Dict, Optional
from abc import ABC, abstractmethod

from ..datasets.base import Dataset
from ..models.base import BaseModel
from .issues import Issue


class Detector(ABC):
    @abstractmethod
    def run(self, model: BaseModel, dataset: Dataset) -> Sequence[Issue]:
        ...


class DetectorRegistry:
    _detectors: Dict[str, Detector] = dict()
    _tags = defaultdict(set)

    @classmethod
    def register(cls, name: str, detector: Detector, tags: Optional[Sequence[str]] = None):
        cls._detectors[name] = detector
        if tags is not None:
            cls._tags[name] = set(tags)

    @classmethod
    def get_detector_classes(cls, tags: Optional[Sequence[str]] = None) -> dict:
        if tags is None:
            return {n: d for n, d in cls._detectors.items()}

        return {n: d for n, d in cls._detectors.items() if cls._tags[n].intersection(tags)}



from typing import Sequence

from .base_detector import BaseTextPerturbationDetector
from .text_transformations import TextTransformation
from ...datasets.base import Dataset
from ...models.base import BaseModel
from ..decorators import detector


@detector(
    name="text_perturbation",
    tags=[
        "text_perturbation",
        "robustness",
        "nlp",
        "classification",
        "regression",
        "generative",
        "text_generation",
        "llm",
    ],
)
class TextPerturbationDetector(BaseTextPerturbationDetector):
    def _get_default_transformations(self, model: BaseModel, dataset: Dataset) -> Sequence[TextTransformation]:
        from .text_transformations import (
            TextUppercase,
            TextLowercase,
            TextTitleCase,
            TextTypoTransformation,
            TextPunctuationRemovalTransformation,
        )

        return [
            TextUppercase,
            TextLowercase,
            TextTitleCase,
            TextTypoTransformation,
            TextPunctuationRemovalTransformation,
        ]



import functools
import inspect
from pathlib import Path
from typing import Optional, List, Union, Type, Callable, Dict

import pandas as pd

from giskard.core.core import DatasetProcessFunctionMeta, DatasetProcessFunctionType
from giskard.core.validation import configured_validate_arguments
from giskard.ml_worker.core.savable import RegistryArtifact
from giskard.ml_worker.testing.registry.decorators_utils import (
    validate_arg_type,
    drop_arg,
    make_all_optional_or_suite_input,
    set_return_type,
)
from giskard.ml_worker.testing.registry.registry import get_object_uuid, tests_registry

SlicingFunctionType = Callable[..., bool]

default_tags = ["filter"]


class SlicingFunction(RegistryArtifact[DatasetProcessFunctionMeta]):
    """
    A slicing function used to subset data.

    Attributes:
        func (SlicingFunctionType): The function used to slice the data.
        row_level (bool): Whether the slicing function should operate on rows or columns.
        cell_level (bool): Whether the slicing function should operate at the cell level.
        params (Dict): Additional parameters for the slicing function.
        is_initialized (bool): Indicates if the slicing function has been initialized.
    """

    func: SlicingFunctionType
    row_level: bool
    cell_level: bool
    params: Dict
    is_initialized: bool

    @classmethod
    def _get_name(cls) -> str:
        """
        Returns the name of the class.

        :return: The name of the class.
        :rtype: str
        """
        return "slices"

    def __init__(self, func: Optional[SlicingFunctionType], row_level=True, cell_level=False):
        """
        Initializes a new instance of the SlicingFunction class.

        :param func: The function used to slice the data.
        :type func: SlicingFunctionType
        :param row_level: Whether the slicing function should operate on rows or the whole dataframe. Defaults to True.
        :type row_level: bool
        """
        self.is_initialized = False
        self.params = {}
        self.func = func
        self.row_level = row_level
        self.cell_level = cell_level

        test_uuid = get_object_uuid(func)
        meta = tests_registry.get_test(test_uuid)
        if meta is None and func is not None:
            meta = DatasetProcessFunctionMeta(func, tags=default_tags, type="SLICE", cell_level=self.cell_level)
            tests_registry.register(meta)
        super().__init__(meta)

    def __call__(self, *args, **kwargs) -> "SlicingFunction":
        self.is_initialized = True
        self.params = kwargs

        for idx, arg in enumerate(args):
            self.params[next(iter([arg.name for arg in self.meta.args.values() if arg.argOrder == idx]))] = arg

        return self

    def execute(self, data: Union[pd.Series, pd.DataFrame]):
        """
        Slices the data using the slicing function.

        Args:
            data (Union[pd.Series, pd.DataFrame]): The data to slice.

        Returns:
            Union[pd.Series, pd.DataFrame]: The sliced data.
        """
        if self.cell_level:
            actual_params = {k: v for k, v in self.params.items() if k != "column_name"}
            return data.loc[data.apply(lambda row: self.func(row[self.params["column_name"]], **actual_params), axis=1)]
        if self.row_level:
            return data.loc[data.apply(lambda row: self.func(row, **self.params), axis=1)]
        else:
            return self.func(data, **self.params)

    @classmethod
    def load(cls, local_dir: Path, uuid: str, meta: DatasetProcessFunctionMeta):
        if meta.process_type == DatasetProcessFunctionType.CODE:
            return super().load(local_dir, uuid, meta)
        else:
            return cls._load_no_code(meta)

    @classmethod
    def _get_meta_class(cls):
        return DatasetProcessFunctionMeta

    @classmethod
    def _load_no_code(cls, meta: DatasetProcessFunctionMeta):
        from ....slicing.slice import Query, QueryBasedSliceFunction

        return QueryBasedSliceFunction(Query.from_clauses(meta.clauses))


def slicing_function(_fn=None, row_level=True, name=None, tags: Optional[List[str]] = None, cell_level=False):
    """
    Decorator that registers a function as a slicing function and returns a SlicingFunction instance.
    It can be used for slicing datasets in a specific way during testing.

    :param _fn: function to decorate. No need to provide this argument, the decorator will automatically take as input the function to decorate.
    :param name: Optional name to use for the function when registering it.
    :param tags: Optional list of tags to use when registering the function.
    :param row_level: Whether to apply the slicing function row-wise (default) or on the full dataframe. If `row_level`
                      is True, the slicing function will receive a row (either a Series or DataFrame),
                      and if False, it will receive the entire dataframe.
    :param cell_level: Whether to apply the slicing function on the cell level. If True, the slicing function
                       will be applied to individual cells instead of rows or the entire dataframe.
    :return: The wrapped function or a new instance of SlicingFunction.
    """

    def inner(func: Union[SlicingFunctionType, Type[SlicingFunction]]) -> SlicingFunction:
        from giskard.ml_worker.testing.registry.registry import tests_registry

        tests_registry.register(
            DatasetProcessFunctionMeta(
                func,
                name=name,
                tags=default_tags if not tags else (default_tags + tags),
                type="SLICE",
                cell_level=cell_level,
            )
        )
        if inspect.isclass(func) and issubclass(func, SlicingFunction):
            return func

        return _wrap_slicing_function(func, row_level, cell_level)()

    if callable(_fn):
        return functools.wraps(_fn)(inner(_fn))
    else:
        return inner


def _wrap_slicing_function(original: Callable, row_level: bool, cell_level: bool):
    slicing_fn = functools.wraps(original)(SlicingFunction(original, row_level, cell_level))

    if not cell_level:
        validate_arg_type(slicing_fn, 0, pd.Series if row_level else pd.DataFrame)

    drop_arg(slicing_fn, 0)

    make_all_optional_or_suite_input(slicing_fn)
    set_return_type(slicing_fn, SlicingFunction)

    return configured_validate_arguments(slicing_fn)



import pandas as pd
from pydantic import validate_arguments
import functools


def configured_validate_arguments(func):
    """
    Decorator to enforce a function args to be compatible with their type hints.
    :return: A wrapper function decorated by pydantic validate_arguments configured to allow arbitrary types check.
    """
    return functools.wraps(func)(validate_arguments(config=dict(arbitrary_types_allowed=True))(func))


def validate_is_pandasdataframe(df):
    assert isinstance(df, pd.DataFrame), "Dataset provided is not a pandas dataframe"



import logging
import os
import time
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import click
import docker
import requests
import yaml
from docker import DockerClient
from docker.errors import NotFound, DockerException
from docker.models.containers import Container
from packaging import version
from packaging.version import InvalidVersion, Version
from tenacity import retry, wait_exponential

import giskard
from giskard.cli_utils import common_options
from giskard.settings import settings
from giskard.utils.analytics_collector import analytics

logger = logging.getLogger(__name__)

giskard_settings_path = settings.home_dir / "server-settings.yml"
IMAGE_NAME = "docker.io/giskardai/giskard"


def create_docker_client() -> DockerClient:
    try:
        return docker.from_env()
    except DockerException as e:
        logger.exception(
            """Failed to connect to Docker. Giskard requires Docker to be installed. If Docker is installed, please run it. Otherwise, please install it.
For an easy installation of Docker you can execute:
- sudo curl -fsSL https://get.docker.com -o get-docker.sh
- sudo sh get-docker.sh""",
            e,
        )
        exit(1)


@click.group(
    "server", help="Giskard UI management", context_settings={"show_default": True}
)
def server() -> None:
    """
    Giskard UI management
    """


def get_version(version=None):
    if version:
        current_settings = _get_settings() or {}
        current_settings["version"] = version
        _write_settings(current_settings)
    else:
        app_settings = _get_settings()
        if app_settings:
            version = app_settings["version"]
        else:
            version = giskard.get_version()
            _write_settings({"version": version})
            latest_version = _fetch_latest_tag()
            message = f"Giskard Server not installed. Using version {version}."
            if latest_version and version != latest_version:
                message += f" Latest available version is {latest_version}. To use it pass a --version argument"
            logger.info(message)
    return version


def get_container_name(version=None):
    if not version:
        version = get_version()
    return f"giskard-server.{version}"


def get_image_name(version=None):
    if not version:
        version = get_version()
    return f"{IMAGE_NAME}:{version}"


def get_container(version=None, quit_if_not_exists=True) -> Optional[Container]:
    name = get_container_name(version)
    try:
        return create_docker_client().containers.get(name)
    except NotFound:
        if quit_if_not_exists:
            logger.error(
                f"Container {name} could not be found. Run `giskard server start` to create the container"
            )
            raise click.Abort()
        else:
            return None


def _is_backend_ready(endpoint) -> bool:
    try:
        response = requests.get(endpoint)
        response.raise_for_status()
        return "UP" == response.json()["status"]
    except KeyboardInterrupt:
        raise click.Abort()
    except BaseException:  # noqa NOSONAR
        return False


def _wait_backend_ready(port: int) -> bool:
    endpoint = f"http://localhost:{port}/management/health"
    backoff_time = 2
    max_duration_second = 3 * 60
    started_time = time.time()
    up = False

    while not up and time.time() - started_time <= max_duration_second:
        time.sleep(backoff_time)
        up = _is_backend_ready(endpoint)
        click.echo(".", nl=False)

    click.echo(".")
    return up


def _start(attached=False, version=None):
    logger.info("Starting Giskard Server")

    settings = _get_settings() or {}
    port = settings.get("port", 19000)
    ml_worker_port = settings.get("ml_worker_port", 40051)

    version = get_version(version)

    _pull_image(version)

    home_volume = _get_home_volume()

    container = get_container(version, quit_if_not_exists=False)

    if not container:
        container = create_docker_client().containers.create(
            get_image_name(version),
            detach=not attached,
            name=get_container_name(version),
            ports={7860: port, 40051: ml_worker_port},
            volumes={home_volume.name: {"bind": "/home/giskard/datadir", "mode": "rw"}},
        )
    container.start()

    up = _wait_backend_ready(port)

    if up:
        logger.info(
            f"Giskard Server {version} started. You can access it at http://localhost:{port}"
        )
    else:
        logger.warning(
            "Giskard backend takes unusually long time to start, "
            "please check the logs with `giskard server logs backend`"
        )


def _check_downloaded(version: str):
    image = get_image_name(version)
    try:
        create_docker_client().images.get(image)
        logger.debug(f"Docker image exists: {image}")
        return True
    except docker.errors.ImageNotFound:
        logger.debug(f"Docker image not found locally: {image}")
        return False


# Version: X.Y.Z
def _pull_image(version):
    if not _check_downloaded(version):
        logger.info(f"Downloading image for version {version}")
        try:
            create_docker_client().images.pull(IMAGE_NAME, tag=version)
        except NotFound:
            logger.error(
                f"Image {get_image_name(version)} not found. Use a valid `--version` argument or check the content of $GSK_HOME/server-settings.yml"
            )
            raise click.Abort()


@retry(wait=wait_exponential(min=0.1, max=5, multiplier=0.1))
def _fetch_latest_tag() -> str:
    """
    Returns: the latest tag from the Docker Hub API. Format: vX.Y.Z
    """
    response = requests.get(
        "https://hub.docker.com/v2/namespaces/giskardai/repositories/giskard/tags?page_size=10"
    )
    response.raise_for_status()
    json_response = response.json()
    latest_tag = "latest"
    latest = next(i for i in json_response["results"] if i["name"] == latest_tag)
    latest_version_image = next(
        i
        for i in json_response["results"]
        if ((i["name"] != latest_tag) and (i["digest"] == latest["digest"]))
    )

    tag = latest_version_image["name"]
    return tag.replace("v", "")


def _write_settings(settings):
    with open(giskard_settings_path, "w") as f:
        yaml.dump(settings, f)


def _get_settings():
    # Check the file exists first
    if not os.path.isfile(giskard_settings_path):
        return None

    return yaml.safe_load(open(giskard_settings_path, "r"))


def _get_home_volume():
    docker_client = create_docker_client()
    try:
        logger.debug("Found existing 'giskard-home' volume, reusing it")
        home_volume = docker_client.volumes.get("giskard-home")
    except NotFound:
        logger.info("Creating a new volume: 'giskard-home'")
        home_volume = docker_client.volumes.create("giskard-home")

    return home_volume


def _expose(token):
    container = get_container()
    if container:
        if container.status != "running":
            print(
                "Error: Giskard server is not running. Please start it using `giskard server start`"
            )
            raise click.Abort()
    else:
        raise click.Abort()
    print("Exposing Giskard Server to the internet...")
    from pyngrok import ngrok
    from pyngrok.conf import PyngrokConfig

    if token:
        ngrok.set_auth_token(token)

    http_tunnel = ngrok.connect(
        19000, "http", pyngrok_config=None if token else PyngrokConfig(region="us")
    )
    tcp_tunnel = ngrok.connect(
        40051, "tcp", pyngrok_config=None if token else PyngrokConfig(region="eu")
    )

    # Only split the last ':' in case the URL contains a port
    tcp_addr = urlparse(tcp_tunnel.public_url)

    print("Giskard Server is now exposed to the internet.")
    print(
        "You can now upload objects to the Giskard Server using the following client: \n"
    )

    print(
        f"""token=...
client = giskard.GiskardClient(\"{http_tunnel.public_url}\", token)

# To run your model with the Giskard Server, execute these three lines on Google Colab:

%env GSK_EXTERNAL_ML_WORKER_HOST={tcp_addr.hostname}
%env GSK_EXTERNAL_ML_WORKER_PORT={tcp_addr.port}
%env GSK_API_KEY=...
!giskard worker start -d -u {http_tunnel.public_url}"""
    )

    ngrok_process = ngrok.get_ngrok_process()
    try:
        # Block until CTRL-C or some other terminating event
        ngrok_process.proc.wait()
    finally:
        print("Shutting down expose.")
        ngrok.kill()


@server.command("start")
@click.option(
    "--attach",
    "-a",
    "attached",
    is_flag=True,
    default=False,
    help="Starts the server and attaches to it, displaying logs in console.",
)
@click.option(
    "--version", "version", required=False, help="Version of Giskard server to start"
)
@common_options
def start(attached, version):
    """\b
    Start Giskard Server.

    By default, the server starts detached and will run in the background.
    You can attach to it by using -a
    """
    analytics.track(
        "giskard-server:start",
        {
            "attached": attached,
            "version": version,
        },
    )
    _start(attached, version)


@server.command("stop")
@common_options
def stop():
    """\b
    Stop Giskard Server.

    Stops a running Giskard server. Does nothing if Giskard server is not running.
    """
    analytics.track("giskard-server:stop")
    container = get_container()
    if container.status != "exited":
        logger.info("Stopping Giskard Server")
        container.stop()
        logger.info("Giskard Server stopped")
    else:
        logger.info(f"Giskard container {container.name} is not running")


@server.command("restart")
@click.argument(
    "service",
    type=click.Choice(["backend", "frontend", "worker", "db"]),
    required=False,
)
@click.option(
    "--hard",
    "hard",
    is_flag=True,
    default=False,
    help="Hard restart. Restarts the whole container",
)
@common_options
def restart(service, hard):
    """\b
    Restart Giskard Server.

    Stops any running Giskard server and starts it again.
    """
    analytics.track(
        "giskard-server:restart",
        {
            "service": service,
            "hard": hard,
        },
    )
    container = get_container()
    if container.status != "running":
        logger.info("Giskard server isn't running")
        _start()
    else:
        if hard:
            logger.info(f"Restarting {container.name} container")
            container.start()
            container.stop()
            if get_container():
                logger.info(f"Container {container.name} restarted")
        else:
            if service:
                logger.info(
                    f"Restarting service {service} in {container.name} container"
                )
                command = (
                    f"supervisorctl -c /opt/giskard/supervisord.conf restart {service}"
                )
            else:
                logger.info(f"Restarting all services in {container.name} container")
                command = "supervisorctl -c /opt/giskard/supervisord.conf restart all"
            for res in container.exec_run(command, stream=True).output:
                print(res.decode())


@server.command("logs")
@click.argument(
    "service",
    type=click.Choice(["backend", "frontend", "worker", "db"]),
    required=False,
)
@click.option(
    "--lines",
    "-l",
    "nb_lines",
    default=300,
    type=click.IntRange(0),
    help="Number of log lines to show",
)
@click.option(
    "--follow",
    "-f",
    "follow",
    is_flag=True,
    default=False,
    help="Follow the logs stream",
)
@common_options
def logs(service, nb_lines, follow):
    """\b
    Prints logs of server services
    """
    analytics.track(
        "giskard-server:logs",
        {
            "service": service,
            "nb_lines": nb_lines,
            "follow": follow,
        },
    )
    container = get_container()
    if not follow:
        if service:
            command = f"tail -{nb_lines} /home/giskard/datadir/run/{service}.log"
        else:
            command = "bash -c 'tail /home/giskard/datadir/run/*.log'"
        print(container.exec_run(command).output.decode())
    else:
        if service:
            command = f"tail -{nb_lines}f /home/giskard/datadir/run/{service}.log"
        else:
            command = "bash -c 'tail -f /home/giskard/datadir/run/*.log'"
        res = container.exec_run(command, stream=True)
        for out in res.output:
            print(out.decode())


@server.command("diagnose")
@click.option(
    "--out_path",
    "-o",
    "local_dir",
    default=os.getcwd(),
    type=click.Path(),
    help="Destination directory to save diagnose archive to",
)
@common_options
def diagnose(local_dir):
    """\b
    Save server logs to a local archive (Useful for support).
    """
    analytics.track("giskard-server:diagnose")
    out_dir = Path(local_dir)
    assert out_dir.is_dir(), "'output' should be an existing directory"
    bits, _ = get_container().get_archive(
        "/home/giskard/datadir/run", encode_stream=True
    )
    from datetime import datetime

    now = datetime.now().strftime("%Y%m%d_%H%M%S_%f")
    out_file = (
        out_dir / f"giskard-diagnose-{get_version().replace('.', '_')}-{now}.tar.gz"
    )
    with open(out_file, "wb") as f:
        for chunk in bits:
            f.write(chunk)
    logger.info(f"Wrote diagnose info to {out_file}")


@server.command("upgrade")
@common_options
@click.argument("version", required=False, default=None)
def upgrade(version):
    """\b
    Update Giskard Server. Uses the latest available version if not specified.
    """
    analytics.track(
        "giskard-server:upgrade",
        {
            "version": version,
        },
    )
    latest_version = _fetch_latest_tag()
    if not version:
        version = latest_version

    installed_version = _get_settings().get("version")
    if installed_version == version:
        logger.info(f"Giskard server is already running version {version}")
        return

    logger.info(f"Updating Giskard Server {installed_version} -> {version}")
    _pull_image(version)
    _write_settings({**_get_settings(), **{"version": version}})
    logger.info(f"Giskard Server upgraded to {version}")


def read_version(version_str: str) -> Version:
    try:
        return version.parse(version_str)
    except InvalidVersion:
        return version.NegativeInfinity


@server.command("status")
@common_options
def status():
    """\b
    Check if server container is running and status of each internal service
    """
    analytics.track("giskard-server:status")
    app_settings = _get_settings()
    if not app_settings:
        logger.info(
            "Giskard Server is not installed. Install using `giskard server start`"
        )
        return
    else:
        version = app_settings["version"]

    logger.info(f"Giskard Server version is set to {version}")

    latest = _fetch_latest_tag()

    if read_version(version) < read_version(latest):
        logger.info(f"A new version is available: {latest}")

    container = get_container()
    if container:
        if container.status == "running":
            logger.info(f"Container {container.name} status:")
            print(
                get_container()
                .exec_run("supervisorctl -c /opt/giskard/supervisord.conf")
                .output.decode()
            )
        else:
            logger.info(
                f"Container {container.name} isn't running ({container.status})"
            )


@server.command("clean")
@click.option(
    "--data", "delete_data", is_flag=True, help="Delete user data (giskard-home volume)"
)
@common_options
def clean(delete_data):
    """\b
    Delete Docker container, container (and possibly a volume) associated with the current version of Giskard Server
    """
    analytics.track("giskard-server:clean", {"delete_data": delete_data})
    data_deletion_confirmed = delete_data and click.confirm(
        "Are you sure you want to delete user data (giskard-home volume)? "
        "This will permanently erase all of the Giskard activity results"
    )

    client = create_docker_client()
    container_name = get_container_name()
    image_name = get_image_name()
    try:
        logger.info(f"Deleting container {container_name}")
        container = client.containers.get(container_name)
        container.stop()
        container.remove()
        logger.info(f"Container {container_name} has been deleted")
    except NotFound:
        logger.info(f"Container {container_name} does not exist")
    try:
        logger.info(f"Deleting image {image_name}")
        client.images.get(image_name).remove(force=True)
        logger.info(f"Image {image_name} has been deleted")
    except NotFound:
        logger.info(f"Image {image_name} does not exist")

    if data_deletion_confirmed:
        try:
            volume = client.volumes.get("giskard-home")
            volume.remove(force=True)
            logger.info("User data has been deleted in 'giskard-home' volume")
        except NotFound:
            logger.info("Volume 'giskard-home' does not exist")


@server.command("expose")
@click.option(
    "--token",
    "token",
    required=False,
    help="In case you have an ngrok account, you can use a token "
    "generated from https://dashboard.ngrok.com/get-started/your-authtoken",
)
@common_options
def expose(token):
    """\b
    Expose your local Giskard Server to the outside world using ngrok to use in notebooks like Google Colab
    """
    analytics.track("giskard-server:expose")
    _expose(token)



import pandas as pd
from typing import List, Optional, Any
from abc import ABC, abstractmethod

from ..datasets import Dataset
from ..models.base import BaseModel


class Issue(ABC):
    group: str = "Other"

    def __init__(
        self,
        model: BaseModel,
        dataset: Dataset,
        level: str,
        info: Optional[Any] = None,
    ):
        self.model = model
        self.dataset = dataset
        self.level = level.lower()
        self.info = info

    def __repr__(self):
        return f"<{self.__class__.__name__} level='{self.level}'>"

    @property
    def is_major(self) -> bool:
        return self.level == "major"

    @property
    def features(self) -> List[str]:
        return []

    @property
    @abstractmethod
    def domain(self) -> str:
        ...

    @property
    @abstractmethod
    def metric(self) -> str:
        ...

    @property
    @abstractmethod
    def deviation(self) -> str:
        ...

    @property
    @abstractmethod
    def description(self) -> str:
        ...

    @abstractmethod
    def examples(self, n=3) -> pd.DataFrame:
        ...

    @property
    @abstractmethod
    def importance(self) -> float:
        ...

    @property
    def slicing_fn(self):
        return None

    @property
    def transformation_fn(self):
        return None

    def generate_tests(self, with_names=False) -> list:
        return []
